
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu
  

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "PLA.h"
#include "MISC_DEFS.h"

int Gemv(int op, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, PLA_Obj beta,
	PLA_Obj y);
int PLA_Gemv_error(int op, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, 
		  PLA_Obj beta, PLA_Obj y);
extern  MPI_Op PLA_SUM;

extern double *timings;



int PLA_Gemv(int op, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, PLA_Obj beta,
	    PLA_Obj y)

/*     A "general" matrix vector multiply. Checks for dimentions
consistency of A, x and y, checks for proper alignment. Then it calls
either PLA_Gemv or PLA_getmv accordingly
 

INPUT: A a m-by-n matrix; 
       op: TRANSPOSE  (=1) or NO_TRANSPOSE (=0) 
       multiscalars alpha and beta distributed amoung the nodes
       multivectors x,y created and properly aligned with A.

OUTPUT: y = alpha * op(A) * x  +  y
*/


{
  int 
    error=0;
  


  error = PLA_Gemv_error(op, alpha, a, x, beta, y);
  
  if (error != 0)
    {
      printf("incorrect input to PLA_Gemv, exiting with error = %d\n", error);
      return error;
    }
  
  error += Gemv(op,alpha,a,x,beta,y);

  return error;
}




int PLA_Gemv_error(int op, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, 
		  PLA_Obj beta, PLA_Obj y)
{
  int
    error = 0,
    x_rowcol, y_rowcol;

  int 
    xtype, ytype,
    idummy,
    m,n,
    im,in,
    x_length, x_width,
    y_length, y_width;
  int
    projection, owner_row, owner_col;

  PLA_Obj_global_length( x, &x_length);
  PLA_Obj_global_length( y, &y_length);
  PLA_Obj_global_width( x, &x_width);
  PLA_Obj_global_width( y, &y_width);

  PLA_Obj_objtype(x, &xtype);
  PLA_Obj_objtype(y, &ytype);


  if( PLA_MATRIX == ytype || 
      PLA_PVECTOR == ytype || 
      PLA_PMVECTOR == ytype)
    {
      PLA_Obj_project_onto( y, &y_rowcol);
      if (PLA_PROJ_ONTO_ROW == y_rowcol)
	{
	  PLA_Obj_global_width(y, &y_length);
	  PLA_Obj_global_length(y, &y_width);
	}
    }
  if( PLA_MATRIX == xtype || 
      PLA_PVECTOR == xtype || 
      PLA_PMVECTOR == xtype)
    {
      PLA_Obj_project_onto( x, &x_rowcol);
      if (PLA_PROJ_ONTO_ROW == x_rowcol)
	{
	  PLA_Obj_global_width(x, &x_length);
	  PLA_Obj_global_length(x, &x_width);
	}
    }


  PLA_Obj_global_info( a, &m, &n, 
                       &projection, &owner_row, &owner_col,
                       &im, &in);

  if (PLA_NO_TRANSPOSE == op)
    {
      if((n != x_length) || (m != y_length))
	{
	  printf("dimensions check failed in PLA_Gemv\n");
	  error++;
	}
	 }
  else if(PLA_TRANSPOSE == op)
	 {
		if((m != x_length) || (n != y_length))
	{
	  printf("dimensions check failed in PLA_Gemv\n");
	  error++;
	}
	 }


  return( error);
}



int Gemv(int op, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, PLA_Obj beta,
	PLA_Obj y)
/* Matrix vector multiply using duplicated panels and PLA_Reduce*/
/* by P.Alpatov                                                */
/*
	NOTE
	Right now PLA_copy cannot handle the Vector to panel case, but can copy
	multivectors to panels. Therefore I temporarily cast incoming vector
	as a multivector......

This routine has been extensivly tested on Hessenberg algorithm and seem
to work fine
*/


{
  int
	 i,j,
	 lld,
	 m, n,
	 in, im,
	 idummy,
	 myrow,mycol,
	 stride,
	 owner_row,
	 owner_col,
	 local_a_length,
	 local_a_width,
	 local_a_lld,
	 ycol_lld;

  void
	 *xtilde_buffer = NULL,
	 *local_buffer = NULL,
	 *local_a_buf = NULL,
	 *alpha_buf = NULL,
	 *beta_buf = NULL;

  PLA_Obj
	 zero = NULL,
	 alpha_tilde = NULL,
	 xtilde = NULL, ytilde = NULL;

  PLA_Template
	 template;

  MPI_Datatype datatype;


  PLA_Obj_template(a, &template);
  PLA_Obj_local_info( a,&local_a_length, &local_a_width,
			&local_a_buf, &idummy, &local_a_lld);
  PLA_Obj_datatype(a, &datatype);


  if (PLA_NO_TRANSPOSE == op){
	 PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW,
					 PLA_ALL_ROWS, &xtilde);
	 PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL,
					 PLA_ALL_COLS, &ytilde);
  }
  else  if ( (PLA_TRANSPOSE == op)  || (PLA_CONJ_TRANS == op )  )
  {
	 PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW,
					 PLA_ALL_ROWS, &ytilde);
	 PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL,
					 PLA_ALL_COLS, &xtilde);
  }
  else
	 printf("Illegal operation in PLA_Gemv.c\n");

  PLA_Copy(x, xtilde);

  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
			  template, &zero);
  PLA_Obj_set_to_zero( zero);

  PLA_Obj_owner_row(alpha, &owner_row);
  PLA_Obj_owner_col(alpha, &owner_col);
  if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS ) {
	 PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS,
					 &alpha_tilde);
	 PLA_Copy( alpha, alpha_tilde );
  }

  PLA_Local_gemv( op, alpha_tilde == NULL ? alpha : alpha_tilde, a, xtilde,
		 zero, ytilde );

  PLA_Scal(beta, y);
  PLA_Reduce_add(ytilde, PLASUM(datatype), y);

  PLA_Obj_free( &ytilde);  PLA_Obj_free( &xtilde);
  PLA_Obj_free( &zero);
  PLA_Obj_free( &alpha_tilde);

  return(0);

}




