

/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"
#include "MISC_DEFS.h"

int Ger(PLA_Obj alpha, PLA_Obj x, PLA_Obj y, PLA_Obj a);
int PLA_Ger_error(PLA_Obj alpha, PLA_Obj x, PLA_Obj y, 
		  PLA_Obj a);






int PLA_Ger(PLA_Obj alpha, PLA_Obj x, PLA_Obj y, PLA_Obj a)

{
  int 
    error=0;
  


  error = PLA_Ger_error(alpha, x, y, a);
  
  if (error != 0)
    {
      printf("incorrect input to PLA_Ger, exiting with error = %d\n", error);
      return error;
    }
  
  error += Ger(alpha,x,y,a);

  return error;
}




int PLA_Ger_error(PLA_Obj alpha, PLA_Obj x, PLA_Obj y, PLA_Obj a)
{
  int
    error = 0,
    x_rowcol, y_rowcol;

  int 
    idummy,
    m,n,
    im,in,
    x_length, x_width,
    y_length, y_width;
  int
    projection, owner_row, owner_col;


  PLA_Obj_global_length( x, &x_length);
  PLA_Obj_global_length( y, &y_length);
  PLA_Obj_global_width( x, &x_width);
  PLA_Obj_global_width( y, &y_width);


  PLA_Obj_global_info( a, &m, &n, 
                       &projection, &owner_row, &owner_col,
                       &im, &in);
  if( !( (((1 == x_length) && (m == x_width)) ||
          ((m == x_length) && (1 == x_width)))
            &&
         (((1 == y_length) && (n == y_width)) ||
          ((n == y_length) && (1 == y_width))) ) )
    {
      printf("dimensions check failed in PLA_Ger\n");
      error++;
    }
 
  return( error);
}



int Ger(PLA_Obj alpha, PLA_Obj x, PLA_Obj y, PLA_Obj a)

{
  int 
    i,j,
    lld, 
    m, n,
    in, im,
    idummy,
    myrow,mycol,
    stride,
    owner_row,
    owner_col,
    local_a_length,
    local_a_width,
    local_a_lld,
    ycol_lld;
  
  void 
    *xtilde_buffer = NULL,
    *local_buffer = NULL,
    *local_a_buf = NULL,
    *alpha_buf = NULL,
    *beta_buf = NULL;
  
  
  PLA_Obj
    alpha_tilde = NULL,
    xtilde = NULL, ytilde = NULL;

  PLA_Template
    template;
  

  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, 
			      PLA_ALL_COLS, &xtilde);
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, 
			      PLA_ALL_ROWS, &ytilde);
  
  PLA_Copy(x, xtilde);
  PLA_Copy(y, ytilde);

  PLA_Obj_owner_row(alpha, &owner_row); 
  PLA_Obj_owner_col(alpha, &owner_col);
  if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS ) {
    PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, 
			       &alpha_tilde);
    PLA_Copy( alpha, alpha_tilde );
  }
  
  PLA_Local_ger( alpha_tilde == NULL ? alpha : alpha_tilde, xtilde,
		 ytilde, a );


  PLA_Obj_free( &ytilde);  PLA_Obj_free( &xtilde);
  PLA_Obj_free( &alpha_tilde);

  return(0);
  
}




