

/***************************************************************************
  11 Febrary 1997 16:02
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/


#include "PLA.h"

int PLA_Syrk ( int uplo, int trans, PLA_Obj alpha, PLA_Obj a,
                                    PLA_Obj beta,  PLA_Obj c )
{
   PLA_Syrk_enter ( uplo, trans, alpha, a, beta, c );

   if ( uplo == PLA_LOW_TRIAN && trans == PLA_NO_TRANS )
      PLA_Syrk_ln ( alpha, a, beta, c );
   else if ( uplo == PLA_LOW_TRIAN && trans == PLA_TRANS    )
      PLA_Syrk_lt ( alpha, a, beta, c );
   else if ( uplo == PLA_UP_TRIAN  && trans == PLA_NO_TRANS )
      PLA_Syrk_un ( alpha, a, beta, c );
   else if ( uplo == PLA_UP_TRIAN  && trans == PLA_TRANS    )
      PLA_Syrk_ut ( alpha, a, beta, c );

   PLA_Syrk_exit ( uplo, trans, alpha, a, beta, c );
   return ( PLA_SUCCESS );
}




int PLA_Syrk_ln ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
   PLA_Syrk_ln_pp ( alpha, a, beta, c );
   return ( PLA_SUCCESS );
}

int PLA_Syrk_ln_mp ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
   int nb_temp, nb_alg, length, size;
   PLA_Template template;
   PLA_Obj a_top = NULL, a_bottom = NULL,
           c_top = NULL, c_bottom = NULL, c_right = NULL;

   PLA_Log_abort ( "PLA_Syrk_ln_mp", "Routine not validated", PLA_ERROR_CODE, 0 );

   PLA_Obj_template ( a, & template );
   PLA_Temp_nb ( template, & nb_temp );
   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb_alg );

   PLA_Obj_view_all ( a, & a_bottom );
   PLA_Obj_view_all ( c, & c_right );

   PLA_Obj_global_length ( a_bottom, & length );

   while ( ( size = min ( nb_alg, length ) ) > 0 )
   {
      PLA_Obj_horz_split_2 ( a_bottom, size, & a_top, & a_bottom );
      PLA_Obj_split_4 ( c_right, size, size, & c_top,    PLA_DUMMY,
                                             & c_bottom, & c_right );

      PLA_Syrk_ln_pp ( alpha, a_top, beta, c_top );

      PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, alpha, a_bottom, a_top, beta, c_bottom );

      PLA_Obj_global_length ( a_bottom, & length );
   }

   PLA_Obj_free ( & a_top );
   PLA_Obj_free ( & a_bottom );
   PLA_Obj_free ( & c_top );
   PLA_Obj_free ( & c_bottom );
   PLA_Obj_free ( & c_right );

   return ( PLA_SUCCESS );
}

int PLA_Syrk_ln_pp ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
   int nb, top_size, top_owner, left_size, left_owner, size, length, width;
   MPI_Datatype datatype;
   PLA_Template template;
   PLA_Obj a_left = NULL, a_right = NULL,
           c_top = NULL,      c_bottom = NULL,     c_right = NULL,
           c2_top = NULL,     c2_bottom = NULL,    c2_right = NULL,
           adup_top = NULL,   adup_bottom = NULL,
           adup_left = NULL,  adup_right = NULL,
           a2dup_top = NULL,  a2dup_bottom = NULL,
           a2dup_left = NULL, a2dup_right = NULL,
           work_col = NULL,  work_row = NULL,    one = NULL;

   PLA_Obj_template ( a, & template );
   PLA_Obj_datatype ( c, & datatype );
   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb );

   PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb, & work_row );
   PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb, & work_col );

   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
   PLA_Obj_set_to_one ( one );

/* scale C by beta */

   PLA_Obj_view_all ( c, & c_right );
   PLA_Obj_global_length ( c_right, & size );
   while ( size > 0 )
   {
      PLA_Obj_vert_split_2 ( c_right, 1, & c_bottom, & c_right );
      PLA_Scal ( beta, c_bottom );
      PLA_Obj_view_shift ( c_right, 1, 0, 0, 0 );
      PLA_Obj_global_length ( c_right, & size );
   }


   PLA_Obj_view_all ( a, & a_right );
   PLA_Obj_global_width ( a_right, & width );

   while ( ( size = min ( nb, width ) ) > 0 )
   {
      PLA_Obj_vert_split_2 ( a_right, size, & a_left, & a_right );
      PLA_Obj_view_all ( work_row, & adup_right );
      PLA_Obj_view_all ( work_col, & adup_bottom );
      PLA_Obj_view_shift ( adup_right, 0, 0, 0, size - nb );
      PLA_Obj_view_shift ( adup_bottom, 0, 0, size - nb, 0 );

      PLA_Obj_set_orientation ( a_left, PLA_PROJ_ONTO_COL );
      PLA_Copy ( a_left, adup_right );
      PLA_Copy ( a_left, adup_bottom );


      PLA_Obj_view_all ( c, & c_right );
      PLA_Obj_global_length ( c_right, & length );

      while ( ( size = min ( nb, length ) ) > 0 )
      {
         PLA_Obj_split_4 ( c_right, size, size, & c_top,    PLA_DUMMY,
                                                & c_bottom, & c_right );
         PLA_Obj_horz_split_2 ( adup_bottom, size, & adup_top, & adup_bottom );
         PLA_Obj_vert_split_2 ( adup_right, size, & adup_left, & adup_right );


         PLA_Obj_view_all ( c_top, & c2_right );
         PLA_Obj_view_all ( adup_top, & a2dup_bottom );
         PLA_Obj_view_all ( adup_left, & a2dup_right );

         PLA_Obj_split_size ( c2_right, PLA_SIDE_TOP, & top_size, & top_owner );
         PLA_Obj_split_size ( c2_right, PLA_SIDE_LEFT, & left_size, & left_owner );

         while ( ( size = min ( top_size, left_size ) ) > 0 )
         {
            PLA_Obj_split_4 ( c2_right, size, size, & c2_top,    PLA_DUMMY,
                                                    & c2_bottom, & c2_right );
            PLA_Obj_horz_split_2 ( a2dup_bottom, size, & a2dup_top, & a2dup_bottom );
            PLA_Obj_vert_split_2 ( a2dup_right, size, & a2dup_left, & a2dup_right );

            PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                             alpha, a2dup_top, one, c2_top );
            PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                             alpha, a2dup_bottom, a2dup_left, one, c2_bottom );

            PLA_Obj_split_size ( c2_right, PLA_SIDE_TOP, & top_size, & top_owner );
            PLA_Obj_split_size ( c2_right, PLA_SIDE_LEFT, & left_size, & left_owner );
         }

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS, alpha, adup_bottom, adup_left,
                                                      one, c_bottom );

         PLA_Obj_global_length ( c_right, & length );
      }

      PLA_Obj_global_width ( a_right, & width );
   }

   PLA_Obj_free ( & a_left );
   PLA_Obj_free ( & a_right );
   PLA_Obj_free ( & adup_top );
   PLA_Obj_free ( & adup_bottom );
   PLA_Obj_free ( & adup_left );
   PLA_Obj_free ( & adup_right );
   PLA_Obj_free ( & a2dup_top );
   PLA_Obj_free ( & a2dup_bottom );
   PLA_Obj_free ( & a2dup_left );
   PLA_Obj_free ( & a2dup_right );
   PLA_Obj_free ( & c_top );
   PLA_Obj_free ( & c_bottom );
   PLA_Obj_free ( & c_right );
   PLA_Obj_free ( & c2_top );
   PLA_Obj_free ( & c2_bottom );
   PLA_Obj_free ( & c2_right );
   PLA_Obj_free ( & work_row );
   PLA_Obj_free ( & work_col );
   PLA_Obj_free ( & one );

   return ( 0 );
}



int PLA_Syrk_lt ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
   PLA_Syrk_lt_pp ( alpha, a, beta, c );
   return ( PLA_SUCCESS );
}

int PLA_Syrk_lt_pp ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
   int nb, top_size, top_owner, left_size, left_owner, size, length;
   MPI_Datatype datatype;
   PLA_Template template;
   PLA_Obj a_top = NULL,     a_bottom = NULL,
           c_top = NULL,     c_bottom = NULL,    c_right = NULL,
           c2_top = NULL,     c2_bottom = NULL,    c2_right = NULL,
           adup_top = NULL,  adup_bottom = NULL,
           adup_left = NULL, adup_right = NULL,
           a2dup_top = NULL,  a2dup_bottom = NULL,
           a2dup_left = NULL, a2dup_right = NULL,
           work_col = NULL,  work_row = NULL,    one = NULL;

   PLA_Obj_template ( a, & template );
   PLA_Obj_datatype ( c, & datatype );
   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb );

   PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb, & work_row );
   PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb, & work_col );

   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
   PLA_Obj_set_to_one ( one );

/* scale C by beta */

   PLA_Obj_view_all ( c, & c_right );
   PLA_Obj_global_length ( c_right, & size );
   while ( size > 0 )
   {
      PLA_Obj_vert_split_2 ( c_right, 1, & c_bottom, & c_right );
      PLA_Scal ( beta, c_bottom );
      PLA_Obj_view_shift ( c_right, 1, 0, 0, 0 );
      PLA_Obj_global_length ( c_right, & size );
   }


   PLA_Obj_view_all ( a, & a_bottom );
   PLA_Obj_global_length ( a_bottom, & length );

   while ( ( size = min ( nb, length ) ) > 0 )
   {
      PLA_Obj_horz_split_2 ( a_bottom, size, & a_top, & a_bottom );
      PLA_Obj_view_all ( work_row, & adup_right );
      PLA_Obj_view_all ( work_col, & adup_bottom );
      PLA_Obj_view_shift ( adup_right, 0, 0, 0, size - nb );
      PLA_Obj_view_shift ( adup_bottom, 0, 0, size - nb, 0 );

      PLA_Obj_set_orientation ( a_top, PLA_PROJ_ONTO_ROW );
      PLA_Copy ( a_top, adup_right );
      PLA_Copy ( a_top, adup_bottom );


      PLA_Obj_view_all ( c, & c_right );
      PLA_Obj_global_length ( c_right, & length );

      while ( ( size = min ( nb, length ) ) > 0 )
      {
         PLA_Obj_split_4 ( c_right, size, size, & c_top,    PLA_DUMMY,
                                                & c_bottom, & c_right );
         PLA_Obj_horz_split_2 ( adup_bottom, size, & adup_top, & adup_bottom );
         PLA_Obj_vert_split_2 ( adup_right, size, & adup_left, & adup_right );


         PLA_Obj_view_all ( c_top, & c2_right );
         PLA_Obj_view_all ( adup_top, & a2dup_bottom );
         PLA_Obj_view_all ( adup_left, & a2dup_right );

         PLA_Obj_split_size ( c2_right, PLA_SIDE_TOP, & top_size, & top_owner );
         PLA_Obj_split_size ( c2_right, PLA_SIDE_LEFT, & left_size, & left_owner );

         while ( ( size = min ( top_size, left_size ) ) > 0 )
         {
            PLA_Obj_split_4 ( c2_right, size, size, & c2_top,    PLA_DUMMY,
                                                    & c2_bottom, & c2_right );
            PLA_Obj_horz_split_2 ( a2dup_bottom, size, & a2dup_top, & a2dup_bottom );
            PLA_Obj_vert_split_2 ( a2dup_right, size, & a2dup_left, & a2dup_right );

            PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                             alpha, a2dup_top, one, c2_top );
            PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                             alpha, a2dup_bottom, a2dup_left, one, c2_bottom );

            PLA_Obj_split_size ( c2_right, PLA_SIDE_TOP, & top_size, & top_owner );
            PLA_Obj_split_size ( c2_right, PLA_SIDE_LEFT, & left_size, & left_owner );
         }

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS, alpha, adup_bottom, adup_left,
                                                      one, c_bottom );

         PLA_Obj_global_length ( c_right, & length );
      }

      PLA_Obj_global_length ( a_bottom, & length );
   }

   PLA_Obj_free ( & a_top );
   PLA_Obj_free ( & a_bottom );
   PLA_Obj_free ( & adup_top );
   PLA_Obj_free ( & adup_bottom );
   PLA_Obj_free ( & adup_left );
   PLA_Obj_free ( & adup_right );
   PLA_Obj_free ( & a2dup_top );
   PLA_Obj_free ( & a2dup_bottom );
   PLA_Obj_free ( & a2dup_left );
   PLA_Obj_free ( & a2dup_right );
   PLA_Obj_free ( & c_top );
   PLA_Obj_free ( & c_bottom );
   PLA_Obj_free ( & c_right );
   PLA_Obj_free ( & c2_top );
   PLA_Obj_free ( & c2_bottom );
   PLA_Obj_free ( & c2_right );
   PLA_Obj_free ( & work_row );
   PLA_Obj_free ( & work_col );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}


int PLA_Syrk_un ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
  int     nb_alg, bottom_size, bottom_owner, size ;
  int     a_right_size, a_right_owner;
  int     right_size, right_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_left = NULL,    a_right = NULL, one = NULL,
					c_topleft = NULL, c_bottomright = NULL, c_topright = NULL,
					work_col = NULL,  a_col_top = NULL,  a_col_bottom = NULL,
					work_row = NULL,  at_row_left = NULL, at_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);

  pla_trap_scal(beta, c, PLA_UPPER_TRIANGULAR);


  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS,1,1,template,&one);
  PLA_Obj_set_to_one ( one );

  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
										 & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
										 & work_col );

  PLA_Obj_view_all ( a, & a_left );

  for ( ; ; )
  {
	 PLA_Obj_global_width(a_left, &a_right_size);
	 if ( (size = min( nb_alg, a_right_size)) == 0) break;
	 PLA_Obj_vert_split_2 ( a_left, -size, & a_left, & a_right );
	 PLA_Obj_vert_split_2 ( work_col,       size,  & a_col_top,   PLA_DUMMY );
	 PLA_Obj_horz_split_2 ( work_row,       size,  & at_row_left, PLA_DUMMY);
	 PLA_Obj_set_orientation ( a_right, PLA_PROJ_ONTO_COL );
	 PLA_Copy ( a_right, a_col_top );
	 PLA_Copy ( a_right, at_row_left );
	 PLA_Obj_view_all ( c, & c_topleft );
	 for( ; ; )
	 {
		PLA_Obj_split_size ( c_topleft, PLA_SIDE_BOTTOM, & bottom_size,
								  & bottom_owner );
		PLA_Obj_split_size ( c_topleft, PLA_SIDE_RIGHT,  & right_size,
								  & right_owner );
		if ( ( size = min3 ( nb_alg, bottom_size, right_size) ) == 0) break;
		PLA_Obj_vert_split_2 (at_row_left, -size, & at_row_left, & at_row_right );
		PLA_Obj_horz_split_2 ( a_col_top, -size, & a_col_top, & a_col_bottom );
		PLA_Obj_split_4 ( c_topleft, -size, -size, & c_topleft,    & c_topright,
																 PLA_DUMMY,      & c_bottomright );
	  if( (my_row == bottom_owner) && (my_col == right_owner) )
			 PLA_Local_syrk ( PLA_UPPER_TRIANGULAR, PLA_NO_TRANSPOSE,
									alpha, a_col_bottom, one, c_bottomright );
		if( (my_col == right_owner) )
	  {
			PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
								  alpha, a_col_top, at_row_right, one, c_topright );
	  }
	 }
  }
  PLA_Obj_free ( & a_left );
  PLA_Obj_free ( & a_right );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_topright );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & at_row_left );
  PLA_Obj_free ( & at_row_right );
  PLA_Obj_free ( & a_col_top );
  PLA_Obj_free ( & a_col_bottom );
  return ( PLA_SUCCESS );
}

/******************************************************************************/
int PLA_Syrk_ut ( PLA_Obj alpha, PLA_Obj a, PLA_Obj beta, PLA_Obj c )
{
  int     nb_alg, bottom_size, bottom_owner, size ;
  int     a_bottom_size, a_bottom_owner;
  int     right_size, right_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_top = NULL,     a_bottom = NULL, one = NULL,
					c_topleft = NULL, c_bottomright = NULL, c_topright = NULL,
					work_col = NULL,  at_col_top = NULL,  at_col_bottom = NULL,
					work_row = NULL,  a_row_left = NULL, a_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);
  pla_trap_scal(beta, c, PLA_UPPER_TRIANGULAR);
  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype,PLA_ALL_ROWS,PLA_ALL_COLS,1, 1,template, &one );
  PLA_Obj_set_to_one ( one );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
										  & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
										  & work_col );
  PLA_Obj_view_all ( a, & a_top );
  for ( ; ; )
  {
	 PLA_Obj_global_length(a_top, &a_bottom_size);
	 if ( (size = min( nb_alg, a_bottom_size)) == 0) break;
	 PLA_Obj_horz_split_2 ( a_top, -size, & a_top, & a_bottom );
	 PLA_Obj_vert_split_2 ( work_col,       size,  &at_col_top, PLA_DUMMY );
	 PLA_Obj_horz_split_2 ( work_row,       size,  &a_row_left, PLA_DUMMY);
	 PLA_Obj_set_orientation ( a_bottom, PLA_PROJ_ONTO_ROW );
	 PLA_Copy ( a_bottom, at_col_top );
	 PLA_Copy ( a_bottom, a_row_left );
	 PLA_Obj_view_all ( c, & c_topleft );
	 for( ; ; )
	 {
		PLA_Obj_split_size ( c_topleft, PLA_SIDE_BOTTOM, & bottom_size,
								  & bottom_owner );
		PLA_Obj_split_size ( c_topleft, PLA_SIDE_RIGHT,  & right_size,
								  & right_owner );
		if ( ( size = min3 ( nb_alg, bottom_size, right_size) ) == 0) break;
		PLA_Obj_vert_split_2 ( a_row_left, -size, & a_row_left, & a_row_right );
		PLA_Obj_horz_split_2 ( at_col_top, -size, & at_col_top, & at_col_bottom );
		PLA_Obj_split_4 ( c_topleft, -size, -size, & c_topleft,    & c_topright,
																 PLA_DUMMY,    &c_bottomright );
	  if( (my_row == bottom_owner) && (my_col == right_owner) )
			 PLA_Local_syrk ( PLA_UPPER_TRIANGULAR, PLA_NO_TRANSPOSE,
									alpha, at_col_bottom, one, c_bottomright);
		if( (my_col == right_owner) )
	  {
			PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
								  alpha, at_col_top, a_row_right, one, c_topright );
	  }
	 }
  }
  PLA_Obj_free ( & a_top );
  PLA_Obj_free ( & a_bottom );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_topright );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & a_row_left );
  PLA_Obj_free ( & a_row_right );
  PLA_Obj_free ( & at_col_top );
  PLA_Obj_free ( & at_col_bottom );
  return ( PLA_SUCCESS );
}

