/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"
#include <stdio.h>

#ifndef min3
#define min3(x, y, z)        min(min(x, y), z)
#endif
int  PLA_Local_trsm(int side, int uplo, int trans, int unit,
                    PLA_Obj alpha, PLA_Obj  A, PLA_Obj B);

/* Note -- resizing the buffers is more complicated in the general case. */
/* There is a simple fix that needs to be in each routine --             */
/* 1) Resize the dup panel to a temp_panel                               */
/* 2) Use the temp panel through the algorithm                           */
/* 3) Do the final split (in the NON-resizing direction) to the original */

int PLA_Trsm_lln(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj  acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,   a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;
  
  int a_sq_top_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_left_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);
                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);

                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_TOP,  &a_sq_top_size,  &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_LEFT, &a_sq_left_size, &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_TOP,  &b_size,         &owner_dont_care);
    if ( ( size = min3( a_sq_top_size, a_sq_left_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_vert_split_2 ( acur, size,              &a1, &a2 );
    PLA_Obj_set_orientation( a1, PLA_PROJ_ONTO_COL );
	PLA_Obj_view_shift(a1tilde,               0,
		                  nb_alg - size,             0,
						                      0);
	PLA_Obj_view_shift(b1tilde,         nb_alg - size,
		                         0,                  0,
								              0);

    PLA_Copy(a1, a1tilde);
    PLA_Obj_horz_split_2 ( a1tilde,   size,         &a1_sq, &a1tilde);
    PLA_Obj_horz_split_2 ( bcur, size,              &b1t, &b2t );
    PLA_Obj_set_orientation( b1t, PLA_PROJ_ONTO_ROW );
/* SOLVE Locally */
    PLA_Local_trsm(PLA_SIDE_LEFT, PLA_LOWER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b1t);
    PLA_Copy(b1t, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, neg_one, a1tilde, b1tilde, one, b2t);
    PLA_Obj_horz_split_2( a2, size, &zeros, &acur);
    PLA_Obj_swap(&bcur, &b2t);
	PLA_Obj_view_shift(a1tilde,               0,
		                  size - nb_alg,             0,
						                      0);
	PLA_Obj_view_shift(b1tilde,          size - nb_alg,
		                         0,                  0,
								              0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);         PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);          PLA_Obj_free(&neg_one);
  PLA_Obj_free(&one);            PLA_Obj_free(&star);
}


int PLA_Trsm_lun(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,   a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_bot_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_right_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);

                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_BOTTOM, &a_sq_bot_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_RIGHT,  &a_sq_right_size, &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_BOTTOM,    &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_bot_size, a_sq_right_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_vert_split_2 ( acur, -size,              &a1, &a2 );
    PLA_Obj_set_orientation( a2, PLA_PROJ_ONTO_COL );

	PLA_Obj_view_shift(a1tilde,               0,
		                  nb_alg - size,             0,
						                      0);
	PLA_Obj_view_shift(b1tilde,         nb_alg - size,
		                         0,                  0,
								              0);
    PLA_Copy(a2, a1tilde);
    PLA_Obj_horz_split_2 ( a1tilde,   -size, &a1tilde, &a1_sq);
    PLA_Obj_horz_split_2 ( bcur,      -size, &b1t, &b2t );
    PLA_Obj_set_orientation( b2t, PLA_PROJ_ONTO_ROW );
/* SOLVE Locally */
    PLA_Local_trsm(PLA_SIDE_LEFT, PLA_UPPER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b2t);
    PLA_Copy(b2t, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, neg_one, a1tilde, b1tilde,
                    one, b1t);
    PLA_Obj_horz_split_2( a1, -size, &acur, &zeros);
    PLA_Obj_swap(&bcur, &b1t);
	PLA_Obj_view_shift(a1tilde,               0,
		                  size - nb_alg,             0,
						  0);
	PLA_Obj_view_shift(b1tilde,          size - nb_alg,
		                         0,                  0,
								              0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);
  PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);
  PLA_Obj_free(&neg_one);
  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}


int PLA_Trsm_rln(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,
           a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_bot_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_right_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_BOTTOM, &a_sq_bot_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_RIGHT,  &a_sq_right_size, &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_RIGHT,  &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_bot_size, a_sq_right_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_horz_split_2 ( acur, -size,              &a1, &a2 );
    PLA_Obj_set_orientation( a2, PLA_PROJ_ONTO_ROW );
	PLA_Obj_view_shift(a1tilde,        nb_alg - size, 
		                          0,                 0,
						                      0);
	PLA_Obj_view_shift(b1tilde,               0,
		                   nb_alg - size,            0,
								              0);
    PLA_Copy(a2, a1tilde);
    PLA_Obj_vert_split_2 ( a1tilde,   -size, &a1tilde, &a1_sq);
    PLA_Obj_vert_split_2 ( bcur,      -size, &b1, &b2 );
    PLA_Obj_set_orientation( b2, PLA_PROJ_ONTO_COL );
/* SOLVE Locally */
    PLA_Local_trsm(PLA_SIDE_RIGHT, PLA_LOWER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b2);
    PLA_Copy(b2, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, neg_one, b1tilde, a1tilde,
                    one, b1);
    PLA_Obj_vert_split_2( a1, -size, &acur, &zeros);
    PLA_Obj_swap(&bcur, &b1);
	PLA_Obj_view_shift(a1tilde,          size - nb_alg, 
		                               0,            0,
						                      0);
	PLA_Obj_view_shift(b1tilde,               0,
		                   size - nb_alg,            0,
								              0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);   PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

int PLA_Trsm_run(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,
           a1_sq = NULL,   a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,
           star = NULL;

  int a_sq_top_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_left_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_TOP,  &a_sq_top_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_LEFT, &a_sq_left_size, &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_LEFT, &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_top_size, a_sq_left_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_horz_split_2   ( acur, size,              &a1, &a2 );
    PLA_Obj_set_orientation( a1, PLA_PROJ_ONTO_ROW );
	PLA_Obj_view_shift(a1tilde,        nb_alg - size,
		                         0,                  0,
								              0);
	PLA_Obj_view_shift(b1tilde,               0,
		                  nb_alg - size,             0,
						                      0);
    PLA_Copy(a1, a1tilde);
    PLA_Obj_vert_split_2 ( a1tilde,   size, &a1_sq, &a1tilde);
    PLA_Obj_vert_split_2 ( bcur,      size, &b1, &b2 );
    PLA_Obj_set_orientation( b1, PLA_PROJ_ONTO_COL );
/* SOLVE Locally */
    PLA_Local_trsm(PLA_SIDE_RIGHT, PLA_UPPER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b1);
    PLA_Copy(b1, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, neg_one, b1tilde, a1tilde,
                    one, b2);
    PLA_Obj_vert_split_2( a2, size, &zeros, &acur);
    PLA_Obj_swap(&bcur, &b2);
	PLA_Obj_view_shift(a1tilde,        size - nb_alg,
		                         0,                  0,
								            0);
	PLA_Obj_view_shift(b1tilde,             0,
		                  size - nb_alg,             0,
						                    0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);   PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

/*********** TRANSPOSE CASES *********************************************/

int PLA_Trsm_llt(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,
           a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_bot_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_right_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_BOTTOM, &a_sq_bot_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_RIGHT,  &a_sq_right_size, &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_BOTTOM, &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_bot_size, a_sq_right_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_horz_split_2   ( acur, -size,          &a1, &a2 );
    PLA_Obj_set_orientation( a2, PLA_PROJ_ONTO_ROW );
	PLA_Obj_view_shift(a1tilde,               0,
		                     nb_alg - size,          0,
						                      0);
	PLA_Obj_view_shift(b1tilde,        nb_alg - size,
		                            0,               0,
								              0);
    PLA_Copy(a2, a1tilde);
    PLA_Obj_horz_split_2 ( a1tilde,   -size, &a1tilde, &a1_sq);
    PLA_Obj_horz_split_2 ( bcur,      -size, &b1, &b2 );
    PLA_Obj_set_orientation( b2, PLA_PROJ_ONTO_ROW );
/* SOLVE Locally -- oddly, UPPER seems to be correct here as does NO_TRANS*/
    PLA_Local_trsm(PLA_SIDE_LEFT, PLA_UPPER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b2);
    PLA_Copy(b2, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS, neg_one, a1tilde, b1tilde,
                    one, b1);
    PLA_Obj_vert_split_2( a1, -size, &acur, &zeros);
    PLA_Obj_swap(&bcur, &b1);
	PLA_Obj_view_shift(a1tilde,               0,
		                  size - nb_alg,             0,
						                      0);
	PLA_Obj_view_shift(b1tilde,          size - nb_alg,
		                       0,                    0,
								              0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);   PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

int PLA_Trsm_lut(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,
           a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_top_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_left_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_TOP,  &a_sq_top_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_LEFT, &a_sq_left_size,  &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_TOP,  &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_top_size, a_sq_left_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_horz_split_2   ( acur, size,          &a1, &a2 );
    PLA_Obj_set_orientation( a1, PLA_PROJ_ONTO_ROW );
	PLA_Obj_view_shift(a1tilde,               0,
		                     nb_alg - size,          0,
						                      0);
	PLA_Obj_view_shift(b1tilde,        nb_alg - size,
		                            0,               0,
								              0);
    PLA_Copy(a1, a1tilde);
    PLA_Obj_horz_split_2 ( a1tilde,   size, &a1_sq, &a1tilde);
    PLA_Obj_horz_split_2 ( bcur,      size, &b1, &b2 );
    PLA_Obj_set_orientation( b1, PLA_PROJ_ONTO_ROW );
/* SOLVE Locally -- oddly, UPPER seems to be correct here as does NO_TRANS*/
    PLA_Local_trsm(PLA_SIDE_LEFT, PLA_LOWER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b1);
    PLA_Copy(b1, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS, neg_one, a1tilde, b1tilde,
                    one, b2);
    PLA_Obj_vert_split_2( a2, size, &zeros, &acur);
    PLA_Obj_swap(&bcur, &b2);
	PLA_Obj_view_shift(a1tilde,              0,
		                  size - nb_alg,                 0,
						                     0);
	PLA_Obj_view_shift(b1tilde,         size - nb_alg,
		                         0,                      0,
								             0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);
  PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

int PLA_Trsm_rlt(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,   a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_top_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_left_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_TOP,  &a_sq_top_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_LEFT, &a_sq_left_size,  &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_LEFT,  &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_top_size, a_sq_left_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_vert_split_2   ( acur, size,          &a1, &a2 );
    PLA_Obj_set_orientation( a1, PLA_PROJ_ONTO_COL );
	PLA_Obj_view_shift(a1tilde,        nb_alg - size, 
		                           0,                    0,
						                      0);
	PLA_Obj_view_shift(b1tilde,               0,
		                       nb_alg - size,          0,
								              0);
    PLA_Copy(a1, a1tilde);
    PLA_Obj_vert_split_2 ( a1tilde,   size, &a1_sq, &a1tilde);
    PLA_Obj_vert_split_2 ( bcur,      size, &b1, &b2 );
    PLA_Obj_set_orientation( b1, PLA_PROJ_ONTO_COL );
/* SOLVE Locally -- oddly, UPPER seems to be correct here as does NO_TRANS*/
    PLA_Local_trsm(PLA_SIDE_RIGHT, PLA_UPPER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b1);
    PLA_Copy(b1, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS, neg_one, b1tilde, a1tilde, 
                    one, b2);
    PLA_Obj_horz_split_2( a2, size, &zeros, &acur);
    PLA_Obj_swap(&bcur, &b2);
	PLA_Obj_view_shift(a1tilde,           size - nb_alg, 
		                            0,                    0,
						                      0);
	PLA_Obj_view_shift(b1tilde,               0,
		                       size - nb_alg,          0,
								              0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);   PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

int PLA_Trsm_rut(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, int unit_nonunit )
{
  PLA_Obj acur = NULL, bcur = NULL,
           a1 = NULL,   b1 = NULL,   neg_one = NULL,
           a2 = NULL,   b2 = NULL,   one = NULL,
           b1t = NULL, b2t = NULL,   a1_sq = NULL, 
           a1tilde = NULL, a1tilde_cur = NULL,   zeros = NULL,
           b1tilde = NULL, b1tilde_cur = NULL,   star = NULL;

  int a_sq_bot_size, owner_dont_care, size, width_a, length_b, nb_alg;
  int a_sq_right_size, b_size;
  double d_one = 1.0, d_neg_one = -1.0;
  PLA_Template  template = NULL;

  /* Scale B by alpha */
  PLA_Scal(alpha, b);

                             /* create constant mscalar 1          */
  PLA_Obj_template(a, &template);
  PLA_Temp_nb(template, &nb_alg);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &neg_one );
  PLA_Obj_set_contents(neg_one, (void*)&d_neg_one);
  PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		       1, 1, template, &one );
  PLA_Obj_set_contents(one, (void*)&d_one);


                             /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );

      /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                nb_alg, &b1tilde );
                             /* Loop until no more of A and B      */
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_split_size(acur, PLA_SIDE_BOTTOM,  &a_sq_bot_size,   &owner_dont_care);
    PLA_Obj_split_size(acur, PLA_SIDE_RIGHT, &a_sq_right_size,  &owner_dont_care);
    PLA_Obj_split_size(bcur, PLA_SIDE_RIGHT,  &b_size,          &owner_dont_care);

    if ( ( size = min3( a_sq_bot_size, a_sq_right_size, b_size )) == 0 ) break;
                             /* split off first column of Acur     */
    PLA_Obj_vert_split_2   ( acur, -size,          &a1, &a2 );
    PLA_Obj_set_orientation( a2, PLA_PROJ_ONTO_COL );
	PLA_Obj_view_shift(a1tilde,       nb_alg - size,
		                          0,                  0,
						                     0);
	PLA_Obj_view_shift(b1tilde,              0,
		                   nb_alg - size,          0,
								             0);
    PLA_Copy(a2, a1tilde);
    PLA_Obj_vert_split_2 ( a1tilde,   -size, &a1tilde, &a1_sq);
    PLA_Obj_vert_split_2 ( bcur,      -size, &b1, &b2 );
    PLA_Obj_set_orientation( b2, PLA_PROJ_ONTO_COL );
/* SOLVE Locally -- oddly, UPPER seems to be correct here as does NO_TRANS*/
    PLA_Local_trsm(PLA_SIDE_RIGHT, PLA_LOWER_TRIANGULAR, PLA_NO_TRANS, unit_nonunit, one, a1_sq, b2);
    PLA_Copy(b2, b1tilde);
                             /* perform local rank-size update     */
    PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS, neg_one, b1tilde, a1tilde, 
                    one, b1);
    PLA_Obj_horz_split_2( a1, -size, &acur, &zeros);
    PLA_Obj_swap(&bcur, &b1);
	PLA_Obj_view_shift(a1tilde,       size - nb_alg,
		                          0,                  0,
						                     0);
	PLA_Obj_view_shift(b1tilde,              0,
		                   size - nb_alg,          0,
								             0);
  }
                             /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde_cur );  PLA_Obj_free( &a1tilde_cur );
  PLA_Obj_free( &b1tilde );      PLA_Obj_free( &a1tilde );
  PLA_Obj_free( &a1_sq);   PLA_Obj_free( &b1t); PLA_Obj_free( &b2t);
  PLA_Obj_free(&zeros);    PLA_Obj_free(&neg_one);  PLA_Obj_free(&one);
  PLA_Obj_free(&star);
}

int  PLA_Local_trsm(int side, int uplo, int trans, int unit,
                    PLA_Obj alpha, PLA_Obj  A, PLA_Obj B)
{
   void *alpha_buf, *a_buf, *b_buf;
   int  length_A, width_A, width_B;
   int  lda, ldb, m, n;
   char  side_char[1], uplo_char[1], trans_char[1], unit_char[1];
   MPI_Datatype datatype;

   PLA_Obj_local_buffer(alpha, &alpha_buf);
   PLA_Obj_local_buffer(A, &a_buf);
   PLA_Obj_local_buffer(B, &b_buf);

   PLA_Obj_datatype(A, &datatype);
 
   PLA_Obj_local_length(B, &m);
   PLA_Obj_local_width (B, &n);
   PLA_Obj_local_ldim(A, &lda);
   PLA_Obj_local_ldim(B, &ldb);
   if(PLA_SIDE_LEFT == side)
   	side_char[0] = 'L';
   else side_char[0] = 'R';
   if(PLA_UPPER_TRIANGULAR == uplo)
   	uplo_char[0] = 'U';
   else uplo_char[0] = 'L';
   if(PLA_NO_TRANSPOSE == trans)
   	trans_char[0] = 'N';
   else trans_char[0] = 'T';
   if(PLA_UNIT_DIAG == unit)
   	unit_char[0] = 'U';
   else unit_char[0] = 'N';
   if(m*n > 0)
   {
   if (MPI_DOUBLE == datatype)
  PLA_dtrsm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            (double*)alpha_buf,
            (double*)a_buf, &lda,
            (double*)b_buf, &ldb);
   else if (MPI_FLOAT == datatype)
  PLA_strsm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            alpha_buf,
            a_buf, &lda,
            b_buf, &ldb);
   else if (MPI_COMPLEX == datatype)
  PLA_ctrsm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            alpha_buf,
            a_buf, &lda,
            b_buf, &ldb);
   else if (MPI_DOUBLE_COMPLEX == datatype)
  PLA_ztrsm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            alpha_buf,
            a_buf, &lda,
            b_buf, &ldb);
   }

}

