#include "mpi.h"
#include "SL.h"
#include "PLA_LAObj_set_local_contents.h"

int  PLA0_LAObj_set_local_contents1( 
                 int            trans, 
                 MPI_Datatype   datatype,
                 void           *buf,
                 int            ld_buf,
                 int            stride_buf,
                 SL_LAObj       laobj )
{
  SL_LAObj_obj_type     objtype;
  MPI_Datatype    datatype_laobj = NULL;
  void           *buf_laobj, *p, *temp_from_buf;
  int             i, j, typesize, stride_laobj, local_length;
  int             local_lld, local_width; 

  SL_LAObj_datatype( laobj, &datatype_laobj );

  if ( datatype != datatype_laobj ) {
    printf("Error in set_local_contents, datatype mismatch\n");
    exit(0);
  }

  SL_LAObj_objtype( laobj, &objtype );
  MPI_Type_size( datatype, &typesize );

  SL_LAObj_buffer( laobj, &buf_laobj );

  if ( objtype == SL_VECTOR ) {
    SL_LAObj_local_stride( laobj, &stride_laobj );
    SL_LAObj_local_length( laobj, &local_length );
    for ( i=0; i<local_length; i++ ) {
      for ( j=0; j<typesize; j++)
      {
	 /* * (char *) buf_laobj++ = * (char *) buf++; */
         *((char*)buf_laobj) = *((char*)buf);
         buf_laobj = (void*) (((char*) buf_laobj) + 1);
         buf       = (void*) (((char*) buf)       + 1);
      }
      /* buf += ( stride_buf-1 ) * typesize; */
      buf = (void*) (((char*)buf) + ( stride_buf-1 ) * typesize);
      /* buf_laobj += ( stride_laobj-1 ) * typesize; */
      buf_laobj = (void*)(((char*)buf_laobj) + ( stride_laobj-1 ) * typesize);
    }
  }
  else if ( objtype == SL_MATRIX ) {
    SL_LAObj_local_lld( laobj, &local_lld );
    SL_LAObj_local_length( laobj, &local_length );
    SL_LAObj_local_width( laobj, &local_width );
    for (j=0; j<local_width; j++ ) {
      /* p = buf_laobj + j*local_lld*typesize; */
      p = (void*) (((char*) buf_laobj) + j*local_lld*typesize);
      /* temp_from_buf = buf + j*ld_buf*typesize; */
      temp_from_buf = (((char*)buf) + j*ld_buf*typesize);
      for (i=0; i<local_length*typesize; i++ )
      {
	/*  * (char *) p++ = * (char *) temp_from_buf++; */
         *((char*)p) = *((char*)temp_from_buf);
         p             = (void*) (((char*)p)             + 1);
         temp_from_buf = (void*) (((char*)temp_from_buf) + 1);
      }
    }
  }
  else{
    printf("Error in set_local_contents, objtype not yet implemented\n");
    exit(0);
  }
}    

    
  

int  PLA0_LAObj_get_local_contents1( 
                 SL_LAObj       laobj,
                 int            trans, 
                 MPI_Datatype   datatype,
                 void           *buf,
                 int            ld_buf,
                 int            stride_buf )
{
  SL_LAObj_obj_type     objtype;

  MPI_Datatype    datatype_laobj = NULL;
  void           *buf_laobj, *p, *temp_from_buf;
  int             i, j, typesize, stride_laobj, local_length;
  int             local_lld, local_width;

  SL_LAObj_datatype( laobj, &datatype_laobj );

  if ( datatype != datatype_laobj ) {
    printf("Error in get_local_contents, datatype mismatch\n");
    exit(0);
  }

  SL_LAObj_objtype( laobj, &objtype );
  MPI_Type_size( datatype, &typesize );

  SL_LAObj_buffer( laobj, &buf_laobj );

  if ( objtype == SL_VECTOR ) {
    SL_LAObj_local_stride( laobj, &stride_laobj );
    SL_LAObj_local_length( laobj, &local_length );
    for ( i=0; i<local_length; i++ ) {
      for ( j=0; j<typesize; j++)
      {
	/* * (char *) buf++ = * (char *) buf_laobj++; */
        *((char*)buf) = *((char*)buf_laobj);
        buf       = (void*) (((char*)buf)       + 1);
        buf_laobj = (void*) (((char*)buf_laobj) + 1);
      }
      /* buf += ( stride_buf-1 ) * typesize; */
      buf = (void*) (((char*)buf) + ( stride_buf-1 ) * typesize);
      /* buf_laobj += ( stride_laobj-1 ) * typesize; */
      buf_laobj = (void*) (((char*) buf_laobj) + ( stride_laobj-1 ) * typesize);
    }
  }
  else if ( objtype == SL_MATRIX ) {
    SL_LAObj_local_lld( laobj, &local_lld );
    SL_LAObj_local_length( laobj, &local_length );
    SL_LAObj_local_width( laobj, &local_width );
    for (j=0; j<local_width; j++ ) {
      /* p = buf_laobj + j*local_lld*typesize; */
      p = (void*) (((char*) buf_laobj) + j*local_lld*typesize);
      /* temp_from_buf = buf + j*ld_buf*typesize; */
      temp_from_buf = (void*) (((char*) buf) + j*ld_buf*typesize);
      for (i=0; i<local_length*typesize; i++ )
      {
 	/*  * (char *) temp_from_buf++ = * (char *) p++; */
        *((char*)temp_from_buf) = *((char*)p);
         temp_from_buf = (void*) (((char*)temp_from_buf) + 1);
         p             = (void*) (((char*)p)             + 1);
      }
    }
  }
  else{
    printf("Error in get_local_contents, objtype not yet implemented\n");
    exit(0);
  }
}    

    
  
