#include "PLA.h"

#define NO_ERROR 0
#define PLA_BAD_OBJ_TYPE         -1011
#define PLA_BAD_DATA_TYPE        -1012
#define PLA_UNMATCHED_DIMENSIONS -1013
#ifndef NULL
#define NULL 0
#endif

#define ROW_ORIENTED -99
#define COL_ORIENTED -98
#define BAD_OBJ_TYPE -897
#define MY_EPSILON 0.1

#ifndef min
#define min(x,y) ( (x) < (y) ? (x) : (y) )
#endif

#define PLA_NO_TRANS       0
#define PLA_NO_TRANSPOSE   0
#define PLA_TRANS          1
#define PLA_TRANSPOSE      1
#define PLA_CONJ_TRANS     2
#define PLA_CONJ_TRANSPOSE 2
#define PLA_CONJUGATE_TRANSPOSE 2
#define PLA_CONJ           3
#define PLA_CONJUGATE      3

#define PLA_SUCCESS                0
#define PLA_ERROR               4740
#define PLA_ERROR_CODE          4740
#define PLA_RETURN_LOG_ERROR    4740
#define PLA_RETURN_LOG_MESSAGE  4740
#define PLA_UNDEFINED           4740

#define UNHANDLED_DUPLICATION   2401
#define UNKNOWN_PROJECTION      2402
#define PLA_UNKNOWN_PROJECTION  2402
#define UNKNOWN_OBJECT_TYPE     2403
#define UNHANDLED_SPLIT_CASE    2404
#define UNIMPLIMENTED_FUNCTION  2405
#define UNKNOWN_SIDE            2406
#define DISALLOWED_OBJECT_TYPE  2407

#define PLA_UNMATCHED_SIZES     2601
#define PLA_UNMATCHED_TYPES     2602
#define PLA_DATATYPE_MISMATCH   2602
#define PLA_INVALID_OBJTYPE     2603
#define PLA_INVALID_DATATYPE    2604
#define PLA_INVALID_TEMPLATE    2605
#define PLA_TEMPLATE_MISMATCH   2606
#define PLA_DIMENSION_MISMATCH  2607
#define PLA_OBJTYPE_MISMATCH    2608
#define PLA_BAD_MALLOC          2609

#define PLA_SIDE_TOP            3500
#define PLA_SIDE_LEFT           3501
#define PLA_SIDE_RIGHT          3502
#define PLA_SIDE_BOTTOM         3503

#define PLA_NO_TRANS       0
#define PLA_NO_TRANSPOSE   0
#define PLA_TRANS         1
#define PLA_TRANSPOSE     1

#define PLA_CHECK_PARAMETER  204
#define PLA_CHECK_PARAMETERS 204
#define PLA_CHECK_SEQUENTIAL 205
#define PLA_INSTRUMENT       206
#define PLA_NO_CHECK_PARAMETER  304
#define PLA_NO_CHECK_PARAMETERS 304
#define PLA_NO_CHECK_SEQUENTIAL 305
#define PLA_NO_INSTRUMENT       306

#define  SYNCHRONOUS 0
#define ASYNCHRONOUS 1
#define  RIGHT  116
#define  LEFT   117
#define  UP     118
#define  DOWN   119
#define  PLA_UP     118
#define  PLA_DOWN   119

#define  PLA_UPPER_TRIANGULAR   401
#define  PLA_LOWER_TRIANGULAR   402
#define  PLA_UP_TRIAN          PLA_UPPER_TRIANGULAR
#define  PLA_LOW_TRIAN         PLA_LOWER_TRIANGULAR
#define  PLA_SHAPE_UPPER_TRAPEZOIDAL 403
#define  PLA_SHAPE_LOWER_TRAPEZOIDAL 404
#define  PLA_SHAPE_UP_TRAP     PLA_SHAPE_UPPER_TRAPEZOIDAL
#define  PLA_SHAPE_LOW_TRAP    PLA_SHAPE_LOWER_TRAPEZOIDAL
#define  PLA_SHAPE_GENERAL      405

#define PLA_NONUNIT_DIAG        410
#define PLA_UNIT_DIAG           411

#define PLA_DIR_UP              450 
#define PLA_DIR_DOWN            451 
#define PLA_DIR_RIGHT           452 
#define PLA_DIR_LEFT            453 
#define PLA_DIR_NOFLOW          454 


/* These are algorythm types used by PLA_Environ_nb_alg
   to specify which algorythm type to determine the 
   optimal blocksize for.  */
#define  PLA_OP_PAN_PAN             501
#define  PLA_OP_PAN_MAT             502
#define  PLA_OP_MAT_PAN             503
#define  PLA_OP_PAN_TRIAN_MAT       504
#define  PLA_OP_SYM_PAN_PAN         505
#define  PLA_OP_TRIAN_MAT_PAN       506
#define  PLA_OP_ALL_ALG             599

extern MPI_Comm PLA_row;
extern MPI_Comm PLA_col;
extern int PLA_me;
extern int PLA_np;
extern int PLA_nprows;
extern int PLA_npcols;
extern int PLA_SYNC;
extern int PLA_flow_direction_x;
extern int PLA_flow_direction_y;




#define  pla_MPI_error(error, function)      \
    if (MPI_SUCCESS != error) {              \
       char buffer[MPI_MAX_ERROR_STRING];    \
       int resultlen;                        \
       printf (" Error returned from " function "\n");  \
       MPI_Error_string (error, buffer, &resultlen);              \
       puts (buffer);                        \
       return(error);                        \
    }
#define PLA_HEAP 112
#define PLA_USER_SUPPLIED 113
#define PLA_USER_BUFFER 113

#define PLA_HAS_SPACE 317
#define PLA_NO_SPACE  318

#define PLA_DEFAULT_OBJ_COUNT 256

#ifdef PNNL
#define pla_malloc(x) pnnl_malloc(x)
#define pla_calloc(x,y) pnnl_calloc(x,y)
#define pla_free(x)   pnnl_free(x)
#endif

#ifndef PLASUM
#define PLASUM(datatype) ( datatype == MPI_DOUBLE_COMPLEX ? PLA_SUM : MPI_SUM)
#endif

