


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/





#ifndef PLA_MATRIX_H
#define PLA_MATRIX_H


#ifndef NULL
#define NULL 0
#endif

/* PLA_Matrix_create
 Creates a matrix.  Subsequently will be freed by PLA_Obj_free.

 This routine sill allocate a buffer large enough to hold the local part of the
 matrix of lenght global_length and datatype.  The template describes how the
 matrix is layed out on the processors, hence how much is on the local processor.
*/
int PLA_Matrix_create (MPI_Datatype      datatype, 
		      int               global_length,
		      int               global_width,
		      PLA_Template       template,
		      int               global_align_row,
		      int               global_align_col,
		      PLA_Obj          *new_matrix);

/* PLA_Matrix_submatrix
 This routines allocates a new matrix object, but does not allocate any memory
 to hold the data.  Instead it just contains a pointer to the buffer allocated
 in PLA_Matrix_create.  Hence changing this buffer will change the origional
 buffer.

 Also once a submatrix is created, freeing the origional matrix will NOT result
 in freeing the submatrix or the buffer origionally allocated.  Instead all memory
 will be preserved untill all submatrixs and the origional matrix have been
 freed with PLA_Obj_free, at which time the master buffer will be freed.
*/
int PLA_Matrix_submatrix (PLA_Obj old_matrix, 
			 int      global_length, 
			 int      global_width,
			 int      matrix_align_row,
			 int      matrix_align_col,
			 PLA_Obj *new_matrix);


int PLA_Matrix_local_all (PLA_Obj          matrix,
			 MPI_Datatype *    datatype,
			 int *             local_length,
			 int *             local_width,
			 void **           buffer,
			 int  *            local_lld);

int PLA_Matrix_global_all (PLA_Obj          matrix,
			  int *             global_length,
			  int *             global_width,
			  int  *            global_align_row,
			  int  *            global_align_col,
			  PLA_Template *     template);


#endif



