


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/





#ifndef PLA_VECTOR_H
#define PLA_VECTOR_H

#ifndef NULL
#define NULL 0
#endif

/*  PLA_Vector_create
 Creates a vector.  Subsequently will be freed by PLA_Obj_free.

 This routine sill allocate a buffer large enough to hold the local part of the
 vector of lenght global_length and datatype.  The template describes how the
 vector is layed out on the processors, hence how much is on the local processor.
 */
int PLA_Vector_create (MPI_Datatype      datatype, 
		      int               global_length,
		      PLA_Template       template,
		      int               global_align,
		      PLA_Obj          *new_vector);

/*  PLA_Vector_subvector
 This routines allocates a new vector object, but does not allocate any memory
 to hold the data.  Instead it just contains a pointer to the buffer allocated
 in PLA_Vector_create.  Hence changing this buffer will change the origional
 buffer.

 Also once a subvector is created, freeing the origional vector will NOT result
 in freeing the subvector or the buffer origionally allocated.  Instead all memory
 will be preserved untill all subvectors and the origional vector have been
 freed with PLA_Obj_free, at which time the master buffer will be freed.
*/
int PLA_Vector_subvector (PLA_Obj        old_vector, 
			 int             global_length, 
			 int             align_in_matrix,
			 PLA_Obj        *new_vector);

/* PLA_Vector_local_all
 Returns useful information about the vector.
*/
int PLA_Vector_local_all (PLA_Obj          vector,
			 MPI_Datatype *    datatype,
			 int *             local_length,
			 void **           buffer,
			 int  *            local_stride);

/* PLA_Vector_global_all
 Returns useful information about the vector.
*/
int PLA_Vector_global_all (PLA_Obj          vector,
			  int *             global_length,
			  int  *            global_align,
			  PLA_Template *     template);


#endif
