/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/




#include "PLA.h"
#include "mpi.h" /* doesn't pla.h do this? */
#include <stdio.h>
#define PLA_TOP PLA_LAOBJ_TOP
#define PLA_LOWER 0
#define PLA_UNIT  1
#define PLA_PVECTOR 6
#define PLA_PMVECTOR 7
#define min3(x,y,z) min(min(x,y), z)
/* Note that these defines SHOULD be in MISC_DEFS.h or PLA.h  */
/* I put them here to get around the compiler problem and to  */
/* point out that they are missing (the defines above that is)*/

int PLA_LU(PLA_LAObj A, PLA_LAObj pivot);
int PLA_Iamax( PLA_LAObj x, PLA_LAObj xmax, PLA_LAObj k );

/**************************************************************************/
/* LU factorization with PARTIAL pivoting.  The unoptimized version.      */
/* Algorithm 3.4.7 from Golub and Van Loan -- 2d Edition                  */
/* The following parts can be optimized once this is thoroughly debugged: */
/* 1) Determination of pivot element (IAMAX) can be optimized to search   */
/*    in parallel instead of "sliding views".                             */
/* 2) The "vertical strip" (nb_alg wide) can be copied to a Mvector.      */
/*    This will allow (1) to occur on all processors.  It will also allow */
/*    column scaling to occur in parallel and puts the data in a form that*/
/*    can be copied more quickly to a duplicated Pmvector -- for the outer*/
/*    product update.                                                     */
/* 3) Rows to be pivoted can be all done at once -- those pieces not in   */
/*    the current vertical strip.  That is, as is described in G&vL --    */
/*    applying the permutation P1 on the rest of A                        */
/*                                                                        */
/* NOTES :                                                                */
/* 1) The no-pivot routine that appears in the manual does not have nb_alg*/
/*    because it uses the current contiguous block size as the square to  */
/*    locally LU-ize.  A similar version should be written for the pivot  */
/*    case.                                                               */
/**************************************************************************/

int PLA_LU(PLA_LAObj A, PLA_LAObj pivot)
/* pivot cannot be a PLA object since its elements are integers?*/
/* It SORT of seems like nb_alg should be a parameter here?     */
{
   PLA_LAObj    a_cur = NULL, A1 = NULL, A1_22 = NULL, temp_row = NULL,
                a11 = NULL, a12 = NULL, a21 = NULL, a22 = NULL,
                a_col = NULL, rest_A1 = NULL, One = NULL, Neg_One = NULL,
                rest_a_col = NULL, rest_a_square_row = NULL, a11_cur = NULL,
                a_top_row = NULL, At = NULL, rest_At = NULL,
                At_rows = NULL, junk_obj = NULL,  a_swap_row = NULL,
                a_col_max = NULL, a_col_max_loc = NULL, a_col_scale = NULL;

   MPI_Datatype a_datatype;
   MPI_Status   status;
   MPI_Comm     Column_communicator;
   int         *pivot_buffer_ptr; /* everyone gets one */

   int          width_A1, global_width_A, global_length_A, local_width_A,
                size_square, nb_alg,count=0,PLA_rank,my_rank_in_column, size_col,i_one=1;
   int          size, a_col_max_loc_val, global_width_a22, global_length_a22;
   int          source_rank, destination_rank, junk_int;
   double       d_one = 1.0, d_neg_one = -1.0;
   float        f_one = 1.0, f_neg_one = -1.0;
   void        *a_col_max_loc_buffer, *scale_buffer;
   void        *temp_buffer_ptr, *receive_buffer_ptr, *send_buffer_ptr;

int row1, col1, row2,col2,row3,col3;
   PLA_Start_num zero_or_one; /* should be an int */

   PLA_Template a_template = NULL;

   PLA_LAObj_view_all(A, & a_cur);
   PLA_LAObj_view_all(A, & At);
   PLA_LAObj_local_width (a_cur, &local_width_A);
   PLA_LAObj_template(A, & a_template);
   PLA_Temp_nb( a_template, &nb_alg);
   PLA_LAObj_datatype(A, &a_datatype);
/* Next line is a DEBUG line -- only uses PLA_rank */
/*   PLA_Temp_comm(a_template, &Column_communicator, &PLA_rank, &size_col); */
   PLA_Temp_comm_col_info(a_template, &Column_communicator, &my_rank_in_column, &size_col);
   PLA_Temp_zero_or_one(a_template, &zero_or_one);
   PLA_Mscalar_create( MPI_INT, PLA_ALL_ROWS, PLA_ALL_COLS,
                       1, 1, a_template, &a_col_max_loc );
   if(MPI_DOUBLE == a_datatype)
   {
      PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		                   1, 1, a_template, &One );
      PLA_LAObj_set_contents(One, (void*)&d_one);
      PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
		                   1, 1, a_template, &Neg_One );
      PLA_LAObj_set_contents(Neg_One, (void*)&d_neg_one);
      PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
                          1, 1, a_template, &a_col_max );
      PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
                          1, 1, a_template, &a_col_scale );
      send_buffer_ptr = (void*)malloc(local_width_A*sizeof(double));
      temp_buffer_ptr = (void*)malloc(local_width_A*sizeof(double));
      receive_buffer_ptr = (void*)malloc(local_width_A*sizeof(double));
   }
   else if(MPI_FLOAT == a_datatype)
   {
      PLA_Mscalar_create( MPI_FLOAT, PLA_ALL_ROWS, PLA_ALL_COLS,
		                   1, 1, a_template, &One );
      PLA_LAObj_set_contents(One, (void*)&f_one);
      PLA_Mscalar_create( MPI_FLOAT, PLA_ALL_ROWS, PLA_ALL_COLS,
		                   1, 1, a_template, &Neg_One );
      PLA_LAObj_set_contents(Neg_One, (void*)&f_neg_one);
      PLA_Mscalar_create( MPI_FLOAT, PLA_ALL_ROWS, PLA_ALL_COLS,
                          1, 1, a_template, &a_col_max );
      PLA_Mscalar_create( MPI_FLOAT, PLA_ALL_ROWS, PLA_ALL_COLS,
                          1, 1, a_template, &a_col_scale );
      send_buffer_ptr = (void*)malloc(local_width_A*sizeof(float));
      temp_buffer_ptr = (void*)malloc(local_width_A*sizeof(float));
      receive_buffer_ptr = (void*)malloc(local_width_A*sizeof(float));
   }

   PLA_LAObj_local_buffer(pivot, &pivot_buffer_ptr);

   while(TRUE)
   {
/* If I don't want to rely on the single parallel parallel TRSM I     */
/* would use the following commented-out splits (as in non-pivot-LU)  */
/*    PLA_LAObj_split_size(a_cur, PLA_SIDE_LEFT, &left_size, &left_owner); */
/*    PLA_LAObj_split_size(a_cur, PLA_TOP,  &top_size,  &top_owner ); */
      PLA_LAObj_global_length(a_cur, &global_length_A);
      PLA_LAObj_global_width (a_cur, &global_width_A);
      size_square = min3(nb_alg, global_length_A, global_width_A);
      if(size_square == 0) break;
/* The next two lines look strange becuase of all of the "junk_obj"s */
/* but I believe that they are correct.                              */
      PLA_LAObj_vert_split_2(a_cur, size_square, &A1,  &junk_obj);
      PLA_LAObj_split_4(a_cur, size_square, size_square, &a11,      &junk_obj,
                                                         &junk_obj, &junk_obj);
      while(TRUE)
      {
         PLA_LAObj_global_width (A1, &width_A1);
         if(width_A1 == 0) break;
         PLA_LAObj_vert_split_2(A1,  1, &a_col,  &rest_A1);
         PLA_LAObj_set_orientation(a_col, PLA_PROJ_ONTO_COL);
         PLA_LAObj_horz_split_2(At,  1, &a_top_row,  &rest_At);
         PLA_LAObj_set_orientation(a_top_row, PLA_PROJ_ONTO_ROW);
         PLA_Iamax(a_col, a_col_max, a_col_max_loc);
         PLA_LAObj_local_buffer(a_col_max_loc, &a_col_max_loc_buffer);
         a_col_max_loc_val = *((int*) a_col_max_loc_buffer);
         *pivot_buffer_ptr = count + a_col_max_loc_val;/*Iamax returns offsets*/
         /* or pivot_buffer_ptr++; */
         *((int*)pivot_buffer_ptr+count) = count + a_col_max_loc_val;

         /* Swap the two rows : This could be done with MPI_Sendrecv()    */
         /* Strip off the top row of one view and do two horz_split_2s to */
         /* get the next one -- the row with the max in it                */
         /* This means that we need a view that is just a horz split on   */
         /* the entire matrix  -- rest_At                                 */
         if(a_col_max_loc_val != zero_or_one) /* if a swap has to be done */
         {
            PLA_LAObj_horz_split_2(At,a_col_max_loc_val-zero_or_one,
                                   &junk_obj,&At_rows);
            PLA_LAObj_horz_split_2(At_rows, 1, &a_swap_row, &junk_obj);
            PLA_LAObj_set_orientation(a_swap_row, PLA_PROJ_ONTO_ROW);
           PLA_LAObj_split_size(a_swap_row,PLA_TOP,&junk_int,&destination_rank);
            PLA_LAObj_split_size(a_top_row, PLA_TOP, &junk_int, &source_rank);
            if(my_rank_in_column == source_rank)
               PLA_LAObj_extract_local_contents(a_top_row, PLA_NO_TRANS,
                                                &i_one, &local_width_A,
                                                send_buffer_ptr, 1, 1);
            if(my_rank_in_column == destination_rank)
               PLA_LAObj_extract_local_contents(a_swap_row, PLA_NO_TRANS,
                                                &i_one, &local_width_A,
                                                receive_buffer_ptr, 1, 1);
/* The row that was the top row should end up in the buffer pointed to by */
/* send_buffer_ptr and the row that had the maxima in it should end up in */
/* the buffer pointed to by receive_buffer_pointer.                       */
            if(destination_rank == source_rank) /* on same processor */
            {
               /* Do nothing */
               if(my_rank_in_column == source_rank)
               {
                  PLA_LAObj_set_local_contents(PLA_NO_TRANS, 1, local_width_A,
                                      send_buffer_ptr, 1, 1, a_swap_row);
                  PLA_LAObj_set_local_contents(PLA_NO_TRANS, 1, local_width_A,
                                      receive_buffer_ptr, 1, 1, a_top_row);
               }
            }
            else
            {
                if(my_rank_in_column == source_rank)
                {
     MPI_Send(send_buffer_ptr, local_width_A, a_datatype, destination_rank,
               99, Column_communicator);
     MPI_Recv(receive_buffer_ptr, local_width_A, a_datatype, destination_rank,
              98, Column_communicator, &status);
                   PLA_LAObj_set_local_contents(PLA_NO_TRANS, 1, local_width_A,
                                          receive_buffer_ptr, 1, 1, a_top_row);
                }
                else if(my_rank_in_column == destination_rank)
                {
     MPI_Recv(send_buffer_ptr, local_width_A, a_datatype, source_rank,
              99, Column_communicator, &status);
     MPI_Send(receive_buffer_ptr, local_width_A, a_datatype, source_rank,
               98, Column_communicator);
                   PLA_LAObj_set_local_contents(PLA_NO_TRANS, 1, local_width_A,
                                          send_buffer_ptr, 1, 1, a_swap_row);
                }
            }
         }
         /* The swap has ALREADY OCCURRED here */
         MPI_Barrier(MPI_COMM_WORLD);
         PLA_LAObj_horz_split_2(a_col, 1, &junk_obj, &rest_a_col);
         PLA_LAObj_set_orientation(rest_a_col, PLA_PROJ_ONTO_COL);
/* While a_col_scale is not strictly needed -- we could just get the inverse */
/* of a_col_max -- it seemed simpler to do this -- but maybe not?            */
         PLA_Local_copy(a_col_max, a_col_scale);
         PLA_LAObj_local_buffer(a_col_scale, &scale_buffer);
         if(a_datatype == MPI_DOUBLE)
         	*((double*) scale_buffer) = 1.0/(*((double*)scale_buffer));
         else if(a_datatype == MPI_FLOAT)
         	*((double*) scale_buffer) = 1.0/(*((float*)scale_buffer));
         PLA_Scal(a_col_scale, rest_a_col); /* No PLA_Scal??? */
/* I am sure there is a better way to do this -- but the first cut should be */
/* the easiest to understand instead of the fastest.                         */
     PLA_LAObj_split_4(a11,1,1,&junk_obj,&rest_a_square_row,&junk_obj,&a11_cur);
         PLA_LAObj_set_orientation(rest_a_square_row, PLA_PROJ_ONTO_ROW);
         PLA_LAObj_split_4(A1, 1, 1, &junk_obj, &junk_obj, &junk_obj, &A1_22);
PLA_Gemm_notransa_notransb_1(1,Neg_One,rest_a_col,rest_a_square_row,One,A1_22);
         PLA_LAObj_view_swap(&a11, &a11_cur);
         PLA_LAObj_view_swap(&A1, &A1_22);
         PLA_LAObj_view_swap(&At, &rest_At);
         count++;
      }
/* Note that a_cur was unchanged in inner loop -- we are essentially stepping */
/* back in our view and then falling forward for the next iteration.          */
      PLA_LAObj_split_4(a_cur,size_square, size_square, &a11, &a12, &a21, &a22);
      PLA_LAObj_set_orientation(a12, PLA_PROJ_ONTO_ROW);
      PLA_LAObj_set_orientation(a21, PLA_PROJ_ONTO_COL);
      /* PLA_Trsm(PLA_SIDE_LEFT, PLA_LOWER, PLA_NO_TRANS, PLA_UNIT, One, a11, a12); */
      PLA_Trsm(One, a11, a12);
MPI_Barrier(MPI_COMM_WORLD);
      PLA_LAObj_global_length(a22, &global_length_a22);
      PLA_LAObj_global_width (a22, &global_width_a22);
/*      if(global_length_a22 * global_width_a22 > 0) */
         PLA_Gemm_notransa_notransb_1(nb_alg, Neg_One, a21, a12, One, a22);
      /* PLA_Gemm(PLA_NO_TRANS, PLA_NO_TRANS, Neg_One, a21, a12, One, a22); */
      PLA_LAObj_view_swap(&a_cur, &a22);
   }
   PLA_LAObj_free(&a_cur);    PLA_LAObj_free(&A1);    PLA_LAObj_free(&temp_row);
   PLA_LAObj_free(&a11);     PLA_LAObj_free(&a12);    PLA_LAObj_free(&a21);
   PLA_LAObj_free(&a22);      PLA_LAObj_free(&rest_a_col);
   PLA_LAObj_free(&a_col);   PLA_LAObj_free(&rest_A1);
   PLA_LAObj_free(&rest_a_square_row); PLA_LAObj_free(&a11_cur);
   PLA_LAObj_free(&a_top_row);   PLA_LAObj_free(&At);  PLA_LAObj_free(&rest_At);
   PLA_LAObj_free(&At_rows); PLA_LAObj_free(&junk_obj);
   PLA_LAObj_free(&a_swap_row);
   PLA_LAObj_free(&a_col_max);   PLA_LAObj_free(&a_col_scale);
   PLA_LAObj_free(&a_col_max_loc); PLA_LAObj_free(&A1_22);
}

