/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/

/* PLA_Local_symm.c */

#include "PLA.h"
#include <stdio.h>

int PLA_Local_symm(int side,
						 int  uplo,
						 PLA_Obj alpha,
						 PLA_Obj A,
						 PLA_Obj B,
						 PLA_Obj beta,
						 PLA_Obj C )
{
	int error=0;
	int local_m, local_n;
	int lda, ldb, ldc;
	MPI_Datatype Datatype;
	void *Abuf, *Bbuf, *Cbuf;
	void *alphabuf, *betabuf;
	char Side_char[1], Uplo_char[1];


	PLA_Obj_local_buffer( A, &Abuf);
	PLA_Obj_local_buffer( B, &Bbuf);
	PLA_Obj_local_buffer( C, &Cbuf);

	PLA_Obj_local_buffer( alpha, &alphabuf);
	PLA_Obj_local_buffer( beta,  &betabuf);

	PLA_Obj_local_length( C, &local_m);
	PLA_Obj_local_width ( C, &local_n);

	if(PLA_SIDE_LEFT == side )
		Side_char[0] = 'L';
	else /* if( PLA_SIDE_RIGHT == side) */
		Side_char[0] = 'R';

	if(PLA_UPPER_TRIANGULAR == uplo)
		Uplo_char[0] = 'U';
	else
		Uplo_char[0] = 'L';

	PLA_Obj_local_ldim ( A, &lda);
	PLA_Obj_local_ldim ( B, &ldb);
	PLA_Obj_local_ldim ( C, &ldc);

	PLA_Obj_datatype(C, &Datatype);

	if (0 != local_m  &&
 	    0 != local_n    )
	{
		if(Datatype == MPI_DOUBLE)
			PLA_dsymm(Side_char, Uplo_char, &local_m, &local_n,
			 alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Datatype == MPI_FLOAT)
		PLA_ssymm(Side_char, Uplo_char, &local_m, &local_n,
			 alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Datatype == MPI_COMPLEX)
			PLA_csymm(Side_char, Uplo_char, &local_m, &local_n,
			 alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Datatype == MPI_DOUBLE_COMPLEX)
			PLA_zsymm(Side_char, Uplo_char, &local_m, &local_n,
			 alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		 else error ++;
	}
return(error);
}
