/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include "mmpi.h"

#define NUM	4000

main( int argc , char * argv[] )
{
  int GTag[ NUM ] ;
  int LTag[ NUM ] ;
  int i, R ;

  /* ---------------------------------------------------------------- */

  if ( MPI_SUCCESS != ( R = MPI_Init( &argc , &argv ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }
  if ( MPI_SUCCESS != ( R = MMPI_Log_init( argv[0] ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }
  if ( MPI_SUCCESS != ( R = MMPI_Enable( MPI_COMM_WORLD ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }

  /* ---------------------------------------------------------------- */

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_get_global\n" );

  for ( i = 0 ; i < NUM ; ++i ) {
    if ( MPI_SUCCESS != ( R = MMPI_Tag_get_global(MPI_COMM_WORLD,GTag + i))) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_verify\n" );

  MMPI_Tag_verify( MPI_COMM_WORLD );

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_rel_global\n" );

  for ( i = 0 ; i < NUM ; i += 2 ) {
    int T = GTag[i] ;
    if ( MPI_SUCCESS != ( R = MMPI_Tag_rel_global(MPI_COMM_WORLD,&T))) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_verify\n" );

  MMPI_Tag_verify( MPI_COMM_WORLD );

  for ( i = 0 ; i < NUM ; i += 2 ) {
    int T ;
    if ( MPI_SUCCESS != ( R = MMPI_Tag_get_global(MPI_COMM_WORLD,&T))) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
    if ( T != GTag[i] ) {
      fprintf(MMPI_Log_file(),"Global tag reservation problem: 0x%x != 0x%x",
         T, GTag[i] );
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_verify\n" );

  MMPI_Tag_verify( MPI_COMM_WORLD );

  /* ---------------------------------------------------------------- */

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_get_local -- %d\n",NUM);

  for ( i = 0 ; i < NUM ; ++i ) {
    if ( MPI_SUCCESS != ( R = MMPI_Tag_get_local( MPI_COMM_WORLD,LTag+i) ) ) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_rel_local\n" );

  for ( i = 0 ; i < NUM ; i += 2 ) {
    int T = LTag[i] ;
    if ( MPI_SUCCESS != ( R = MMPI_Tag_rel_local( MPI_COMM_WORLD, &T ) ) ) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  fprintf(MMPI_Log_file(),"main: MMPI_Tag_get_local\n" );

  for ( i = 0 ; i < NUM ; i += 2 ) {
    int T ;
    if ( MPI_SUCCESS != ( R = MMPI_Tag_get_local( MPI_COMM_WORLD, &T ) ) ) {
      MPI_Abort( MPI_COMM_WORLD , R );
    }
    if ( T != LTag[i] ) {
      fprintf(MMPI_Log_file(),"Local tag reservation problem: 0x%x != 0x%x",
         T, LTag[i] );
      MPI_Abort( MPI_COMM_WORLD , R );
    }
  }

  /* ---------------------------------------------------------------- */

  MMPI_Disable( MPI_COMM_WORLD );
  MPI_Finalize();

  return 0 ;
}

