
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/
#include "PLA.h"
/* SHOULD the buffer size be a long???? */
int PLA_Vector_buf_size (MPI_Datatype      datatype,
		      int               global_length,
		      PLA_Template       template,
		      int               global_align,
		      int              *bytes_required)
{
   return(PLA_Mvector_buf_size(datatype, global_length, 1, template,
                               global_align, bytes_required) );
}

/* Buffer flags
1 : A buffer is passed for use but the object does not exist
0 : A buffer is not passed and one should be malloced
-1: A buffer is not passed and one should NOT be malloced
*/
int PLA_Vector_create (MPI_Datatype      datatype,
		      int               global_length,
		      PLA_Template       template,
		      int               global_align,
		      PLA_Obj          *new_vector)
{

  int           error=0;
  error=PLA_Mvector_create (datatype,
                            global_length,
                            1,
                            template,
                            global_align,
                            new_vector);
  (*new_vector)->obj_type = PLA_VECTOR;
   /* new fields */
  (*new_vector)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
  (*new_vector)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
}

int PLA_Vector_create_buf (MPI_Datatype      datatype,
		      int               global_length,
		      PLA_Template      template,
		      int               global_align,
			   int               buffer_size,
			   void             *data_buffer,
		      PLA_Obj          *new_vector)
{
   int  buffer_flag = 1; /* indicates that a buffer is passed for use */
   return PLA_Mvector_create_base(datatype, global_length, 1, template,
	                               global_align,
                                	 buffer_flag, buffer_size, data_buffer,
								          new_vector);
   (*new_vector)->obj_type = PLA_VECTOR;
   (*new_vector)->allocation_type = PLA_USER_BUFFER;    /* vs. PLA_HEAP */
   (*new_vector)->has_data_space  = PLA_HAS_SPACE;      /* vs. PLA_NO_SPACE */
}

int PLA_Vector_create_no_buf (MPI_Datatype      datatype,
		      int               global_length,
		      PLA_Template      template,
		      int               global_align,
		      PLA_Obj          *new_vector)
{
   int  buffer_flag = -1; /* indicates that a buffer is passed for use */
   void *data_buffer = NULL;
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   return PLA_Mvector_create_base(datatype, global_length,1,template,
	                               global_align,
                                  buffer_flag, buffer_size, data_buffer,
								          new_vector);
   (*new_vector)->obj_type = PLA_VECTOR;
   (*new_vector)->allocation_type = PLA_UNDEFINED;    /* not defined until space attached  */
   (*new_vector)->has_data_space  = PLA_NO_SPACE;    /* vs. PLA_HAS_SPACE */
}

int PLA_Vector_attach_buf(PLA_Obj vector, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(vector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = vector->local_length * vector->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Vector_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else vector->master_buffer   = data_buffer;
   vector->allocation_type = PLA_USER_SUPPLIED;
   vector->has_data_space  = PLA_HAS_SPACE;
}

int PLA_Vector_attach(PLA_Obj vector)
{
   /* Need to modularize the matrix create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(vector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = vector->local_length * vector->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      vector->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (vector->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Vector_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else vector->master_buffer = NULL;
   vector->allocation_type    = PLA_HEAP;
   vector->has_data_space     = PLA_HAS_SPACE;
}



int PLA_Vector_subvector (PLA_Obj old_vector, 
			 int      global_length, 
			 int      align_in_vector,
			 PLA_Obj *new_vector)
{
  
  
  PLA_Obj      vector;
  PLA_Template   template;
  int           global_align_row, local_len;
  int           local_align;
  PLA_Start_num  zero_or_one;
  int           data_size; 
  int           error=0;
  int           index;
  
  if ( old_vector->obj_type != PLA_VECTOR) {
    printf ("Invalid object passed to PLA_Vector_subvector: NonVector\n");
    return (1);
  }

  if (*new_vector != NULL) error += PLA_Obj_free (new_vector);

  template = old_vector->template;
  error += PLA_Temp_zero_or_one (template, &zero_or_one);
 
  index = PLA_Get_free_object_struct();
  if(-1 == index) /* Couldn't get a free one */
  {
     printf("Failure to get obj_struct in vector_subvector\n");
     return(1);
  } 
  vector = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]); 
  /* 
     By default copy all the fields in one blow.  Then modify
     the ones that need modifying.
     */
  
  *vector = *old_vector;

  if (global_length == PLA_DIM_ALL) global_length = vector->global_length;
  else                             vector->global_length = global_length;

  if (align_in_vector == PLA_ALIGN_FIRST) align_in_vector = zero_or_one;

  global_align_row         = 
    old_vector->global_align_row + align_in_vector - (int)zero_or_one;
  vector->global_align_row = global_align_row;

  error += PLA_Temp_vector_local_length (template,
					global_length, 
					global_align_row,
					&local_len);
  if (error) return(error);

  vector->local_length = local_len;

  /*
     Determine how much of the local vector to skip at the beginning.
     
     There is a detailed description of this in the PLA_Matrix_submatrix
     routine.  See details there.
     */
  if (local_len!=0) {

    align_in_vector -= (int) zero_or_one;
    error = PLA_Temp_vector_local_length (template,
					 align_in_vector,
					 old_vector->global_align_row,
					 &local_align);
    
    MPI_Type_size (vector->datatype, &data_size);
    vector->buffer = (void *)
      ((char *)(vector->buffer) + data_size * local_align);
  }
  else
    vector->buffer = NULL;

 
  /* 
     increase the number of views.  vector->views is a pointer to an
     integer stored in all the data structures that use this master
     buffer.   We don't want to deallocate the master buffer as long
     as there is a subbuffer pointing to it.
     */
  
  (*(vector->views))++;

  /* Set the first and last block size in both dimentions */
  
  error += PLA_Temp_vector_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(vector->owner_top_block),
					     &(vector->length_top_block));
  error += PLA_Temp_vector_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(vector->owner_bottom_block),
					     &(vector->length_bottom_block));
  vector->width_left_block = 0;
  vector->width_right_block  = 0;
  vector->owner_left_block = 0;
  vector->owner_right_block  = 0;
  
  *new_vector = vector;
  return(error);
}




int PLA_Vector_local_all (PLA_Obj          vector,
			 MPI_Datatype *    datatype,
			 int *             local_length,
			 void **           buffer,
			 int  *            local_stride)
{
  int error=0;

  error += PLA_Obj_datatype       (vector,  datatype);
  error += PLA_Obj_local_length   (vector,  local_length);
  error += PLA_Obj_local_buffer         (vector,  buffer); 
  error += PLA_Obj_local_stride   (vector,  local_stride); 
  return (error);
}




int PLA_Vector_view_shift (PLA_Obj vector, 
			   int     length_top,
			   int     width_left,
			   int     width_right,
			   int     length_bottom)

{
  int error = 0;

  error += PLA_Mvector_view_shift( vector, length_top, width_left,
				           width_right, length_bottom);
  
  return(error);
}


int PLA_Vector_global_all (PLA_Obj          vector,
			  int *             global_length,
			  int  *            global_align,
			  PLA_Template *     template)
     
{
  int error=0;

  error += PLA_Obj_global_length            (vector,  global_length);
  error += PLA_Obj_global_align_row  (vector,  global_align); 
  error += PLA_Obj_template          (vector,  template); 
  return (error);
}




