#include "PLA.h"
#include "POOCLA.h"

int PLA_test_ooc_dtrsm ( int my_pe, PLA_Template template )
{
  int n, m, np, lda_all, ldb_all;
  double alpha, time, mflops, *a_all = NULL, *b_all = NULL, *c_all = NULL;
  PLA_Obj a_obj=NULL, b_obj=NULL, alpha_obj = NULL; 
  POOCLA_Obj a_ooc = NULL, b_ooc = NULL;
  MPI_Comm comm1d;


  PLA_Temp_comm1D( template, &comm1d );
  MPI_Comm_size( comm1d, &np );

  if ( my_pe == 0 )
  {
     printf ( "m n\n" );
     scanf ( "%d%d", & m, & n );
  }

  MPI_Bcast ( & m, 1, MPI_INT, 0, comm1d );
  MPI_Bcast ( & n, 1, MPI_INT, 0, comm1d );


   PLA_Mscalar_create(MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,&alpha_obj);
   PLA_Obj_set_to_one(alpha_obj);
   alpha=1.0e+0;

   a_all = (double *) pla_calloc( m * m, sizeof( double ) );
   lda_all = m;
   if ( my_pe == 0 ) PLA_test_dgenmatrix( "D", 10.0, m, m, a_all, lda_all );
   MPI_Bcast ( a_all, m * m , MPI_DOUBLE, 0, comm1d );

   POOCLA_create ( template, MPI_DOUBLE, m, m, 0, 0,
                    PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, "Mat_A", & a_ooc );
   POOCLA_create_PLA_Obj_conf_to ( a_ooc, PLA_MATRIX, & a_obj );
   PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
   POOCLA_write ( a_ooc, a_obj );
   POOCLA_validate_filestatus ( a_ooc );


   b_all = (double *) pla_calloc( m * n, sizeof( double ) );
   ldb_all = m;
   if ( my_pe == 0 ) PLA_test_dgenmatrix( "G", 1000.0000, m, n, b_all, ldb_all );
   MPI_Bcast ( b_all, m * n, MPI_DOUBLE, 0, comm1d );

   PLA_dtrmm( "L", "L", "T", "N", &m, &n, &alpha, a_all, &lda_all, b_all, &ldb_all);

   POOCLA_create ( template, MPI_DOUBLE, m, n, 0, 0,
                    PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, "Mat_B", & b_ooc );
   POOCLA_create_PLA_Obj_conf_to ( b_ooc, PLA_MATRIX, & b_obj );
   PLA_Test_gdist_to_local( b_all, ldb_all, b_obj );
   POOCLA_write ( b_ooc, b_obj );
   POOCLA_validate_filestatus ( b_ooc );




   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime();
   PLA_dtrsm( "L", "L", "T", "N", &m, &n, &alpha, a_all, &lda_all, b_all, &ldb_all); 
   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime() - time;
   mflops = 1.0e-6 * m * n * m / ( 2.0e+0 * time );

   if ( my_pe == 0 ) printf ( "Serial : %f seconds @ %f mflops\n", time, mflops );


   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime();
   PLA_Trsm ( PLA_SIDE_LEFT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
              alpha_obj, a_obj, b_obj );

   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime() - time;
   mflops = 1.0e-6 * m * n * m / ( 2.0e+0 * time * np );
   if ( my_pe == 0 )
      printf ( "Incore : %f seconds @ %f mflops per processor\n", time, mflops );

   POOCLA_activate_instrumentation ();
   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime();
   POOCLA_Trsm ( PLA_SIDE_LEFT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                  alpha_obj, a_ooc, b_ooc );
   MPI_Barrier( MPI_COMM_WORLD );
   time = MPI_Wtime() - time;
   mflops = 1.0e-6 * m * n * m / ( 2.0e+0 * time * np );
   if ( my_pe == 0 )
      printf ( "Outcore : %f seconds @ %f mflops per processor\n", time, mflops );
   POOCLA_read_time ( & mflops );
   MPI_Reduce ( & mflops, & time, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
   if ( my_pe == 0 )
      printf ( "Average Read Time : %f seconds\n", time / np );
   POOCLA_write_time ( & mflops );
   MPI_Reduce ( & mflops, & time, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
   if ( my_pe == 0 )
      printf ( "Average Write Time : %f seconds\n", time / np );

   POOCLA_read ( b_ooc, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & a_obj );

   
   PLA_test_compare( b_all, ldb_all, b_obj);
   PLA_test_compare( b_all, ldb_all, a_obj);


   if ( a_all != NULL ) pla_free( a_all );
   if ( b_all != NULL ) pla_free( b_all );

   PLA_Obj_free( &a_obj );
   PLA_Obj_free( &b_obj);
   PLA_Obj_free( &alpha_obj);

   POOCLA_free ( & a_ooc );
   POOCLA_free ( & b_ooc );
}
