#include "PLA.h"

/*
 * PLAPACK Level-2 BLAS Implementation
 */
int PLA_opt_CHOLxR1 ( int int_uplo, PLA_Obj a )
{
   int size;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_col = NULL, a_next = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_global_length ( a_next, & size );

      if ( size == 0 ) break;

      PLA_Obj_split_4 ( a_next, 1, 1, & a_cur, PLA_DUMMY,
                                      & a_col, & a_next );

      PLA_Take_sqrt ( a_cur );

      PLA_Obj_objtype_cast ( a_cur, PLA_MSCALAR );
      PLA_Inv_scal ( a_cur, a_col );

      PLA_Syr ( PLA_LOW_TRIAN, min_one, a_col, a_next );
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation
 */
int PLA_opt_CHOLxR2 ( int int_uplo, PLA_Obj a )
{
   int size, nb_alg;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_col = NULL, a_next = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_global_length ( a_next, & size );

      if ( ( size = min ( size, nb_alg ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_opt_CHOLxR1 ( PLA_LOW_TRIAN, a_cur );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a_cur, a_col );

      PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, a_col, one, a_next );
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation (Local_Chol Optimization)
 */
int PLA_opt_CHOLxR3 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_col = NULL, a_next = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_split_size ( a_next, PLA_SIDE_TOP, & size_top, & owner_top );
      PLA_Obj_split_size ( a_next, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_Local_chol ( a_cur );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a_cur, a_col );

      PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, a_col, one, a_next );
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation (Inline Trsm)
 */
int PLA_opt_CHOLxR4 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_cur_dup = NULL, a_col = NULL, a_next = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_split_size ( a_next, PLA_SIDE_TOP, & size_top, & owner_top );
      PLA_Obj_split_size ( a_next, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_Local_chol ( a_cur );

/*      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_INHERIT, & a_cur_dup );*/
      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_ALL_COLS, & a_cur_dup );

      PLA_Copy ( a_cur, a_cur_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a_cur_dup, a_col );

      PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, a_col, one, a_next );
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_cur_dup );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation (Inline Syrk)
 */
int PLA_opt_CHOLxR5 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_cur_dup = NULL,
           a_col = NULL, a_col_dup = NULL, a_row_dup = NULL,
           a_next = NULL, a_rest = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_split_size ( a_next, PLA_SIDE_TOP, & size_top, & owner_top );
      PLA_Obj_split_size ( a_next, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_Local_chol ( a_cur );

/*      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_INHERIT, & a_cur_dup );*/
      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_ALL_COLS, & a_cur_dup );

      PLA_Copy ( a_cur, a_cur_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a_cur_dup, a_col );

      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    size, & a_col_dup );
      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    size, & a_row_dup );
      PLA_Obj_set_orientation ( a_col, PLA_PROJ_ONTO_COL );

      PLA_Copy ( a_col, a_col_dup );
      PLA_Copy ( a_col, a_row_dup );

      pla_syrk_opt_local_update ( min_one, a_col_dup, a_row_dup, one, a_next );

/*      PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
      PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
      PLA_Obj_vert_split_2 ( a_row_dup, 0, & a_row_dup, PLA_DUMMY );

      PLA_Obj_view_all ( a_next, & a_rest );

      while ( TRUE )
      {
         PLA_Obj_split_size ( a_rest, PLA_SIDE_TOP, & size_top, & owner_top );
         PLA_Obj_split_size ( a_rest, PLA_SIDE_LEFT, & size_left, & owner_left );

         if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

         PLA_Obj_view_shift ( a_cur, 0, 0, size, size );
         PLA_Obj_view_shift ( a_col, size, 0, size, 0 );
         PLA_Obj_view_shift ( a_col_dup, size, 0, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, 0, size, 0 );

         PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_TRANS,
                          min_one, a_row_dup, one, a_cur );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          min_one, a_col_dup, a_row_dup, one, a_col );

         PLA_Obj_view_shift ( a_rest, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, size, 0, 0 );
      }
*/
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_cur_dup );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_col_dup );
   PLA_Obj_free ( & a_row_dup );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & a_rest );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation (uses nb_alg)
 */
int PLA_opt_CHOLxR6 ( int int_uplo, PLA_Obj a )
{
   int nb_alg, size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_cur_dup = NULL,
           a_col = NULL, a_col_dup = NULL, a_col_mv = NULL, a_row_dup = NULL,
           a_next = NULL, a_rest = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_global_length ( a_next, & size );

      if ( ( size = min ( size, nb_alg ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_opt_CHOLxR5 ( int_uplo, a_cur );

/* I would love to in-line trsm as describe below, however for alignments != zero
   the matrix->mscalar copy produces a core dump.  For now, settle for global
   blas call */

/* begin inlined code (commented out) */
/*      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_INHERIT, & a_cur_dup );*/
/*      PLA_Mscalar_create_conf_to ( a_cur, PLA_ALL_ROWS, PLA_ALL_COLS, & a_cur_dup );

      PLA_Mvector_create_conf_to ( a_col, PLA_INHERIT, & a_col_mv );

      PLA_Copy ( a_cur, a_cur_dup );
      PLA_Copy ( a_col, a_col_mv );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a_cur_dup, a_col_mv );

      PLA_Copy ( a_col_mv, a_col );

      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    size, & a_col_dup );
      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    size, & a_row_dup );
      PLA_Obj_set_orientation ( a_col, PLA_PROJ_ONTO_COL );

      PLA_Copy ( a_col_mv, a_row_dup );
      PLA_Copy ( a_col, a_col_dup );
*/
/* end inlined code (commented out) */

/* begin non-inlined code */
      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a_cur, a_col );
      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    size, & a_col_dup );
      PLA_Pmvector_create_conf_to ( a_next, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    size, & a_row_dup );
      PLA_Obj_set_orientation ( a_col, PLA_PROJ_ONTO_COL );
      PLA_Copy ( a_col, a_row_dup );
      PLA_Copy ( a_col, a_col_dup );
/* end non-inlined code */


/*      pla_syrk_opt_local_update ( min_one, a_col_dup, a_row_dup, one, a_next );*/

      PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
      PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
      PLA_Obj_vert_split_2 ( a_row_dup, 0, & a_row_dup, PLA_DUMMY );

      PLA_Obj_view_all ( a_next, & a_rest );

      while ( TRUE )
      {
         PLA_Obj_split_size ( a_rest, PLA_SIDE_TOP, & size_top, & owner_top );
         PLA_Obj_split_size ( a_rest, PLA_SIDE_LEFT, & size_left, & owner_left );

         if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

         PLA_Obj_view_shift ( a_cur, 0, 0, size, size );
         PLA_Obj_view_shift ( a_col, size, 0, size, 0 );
         PLA_Obj_view_shift ( a_col_dup, size, 0, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, 0, size, 0 );

         PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_TRANS,
                          min_one, a_row_dup, one, a_cur );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          min_one, a_col_dup, a_row_dup, one, a_col );

         PLA_Obj_view_shift ( a_rest, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, size, 0, 0 );
      }
   }

   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_cur_dup );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_col_dup );
   PLA_Obj_free ( & a_col_mv );
   PLA_Obj_free ( & a_row_dup );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & a_rest );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 Global BLAS Implementation (preallocs buffers)
 */
int PLA_opt_CHOLxR7 ( int int_uplo, PLA_Obj a )
{
   int nb_alg, size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj a_cur = NULL, a_cur_dup = NULL,
           a_col = NULL, a_col_dup = NULL, a_col_mv = NULL, a_row_dup = NULL,
           buf_a_cur_dup = NULL, buf_a_col_dup = NULL, buf_a_col_mv = NULL,
           buf_a_row_dup = NULL,
           a_next = NULL, a_rest = NULL,
           one = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, nb_alg, nb_alg,
                        template, & buf_a_cur_dup );

   PLA_Mvector_create_conf_to ( a, nb_alg, & buf_a_col_mv );

   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                 nb_alg, & buf_a_col_dup );
   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                 nb_alg, & buf_a_row_dup );

   PLA_Obj_view_all ( a, & a_next );

   while ( TRUE )
   {
      PLA_Obj_global_length ( a_next, & size );

      if ( ( size = min ( size, nb_alg ) ) == 0 ) break;

      PLA_Obj_split_4 ( a_next, size, size, & a_cur, PLA_DUMMY,
                                            & a_col, & a_next );

      PLA_Obj_view_shift ( buf_a_col_mv, size, 0, 0, 0 );
      PLA_Obj_view_shift ( buf_a_col_dup, size, 0, 0, 0 );
      PLA_Obj_view_shift ( buf_a_row_dup, 0, size, 0, 0 );

      PLA_opt_CHOLxR5 ( int_uplo, a_cur );

/* I would love to in-line trsm as describe below, however for alignments != zero
   the matrix->mscalar copy produces a core dump.  For now, settle for global
   blas call */

/* begin inlined code (commented out) */
/*      PLA_Obj_view_all ( buf_a_cur_dup, & a_cur_dup );
      PLA_Obj_view_shift ( a_cur_dup, 0, 0, size - nb_alg, size - nb_alg );

      PLA_Obj_view_all ( buf_a_col_mv, & a_col_mv );
      PLA_Obj_view_shift ( a_col_mv, 0, 0, size - nb_alg, 0 );

      PLA_Copy ( a_cur, a_cur_dup );
      PLA_Copy ( a_col, a_col_mv );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a_cur_dup, a_col_mv );

      PLA_Copy ( a_col_mv, a_col );

      PLA_Obj_view_all ( buf_a_col_dup, & a_col_dup );
      PLA_Obj_view_shift ( a_col_dup, 0, 0, size - nb_alg, 0 );

      PLA_Obj_view_all ( buf_a_row_dup, & a_row_dup );
      PLA_Obj_view_shift ( a_row_dup, 0, 0, 0, size - nb_alg );

      PLA_Obj_set_orientation ( a_col, PLA_PROJ_ONTO_COL );

      PLA_Copy ( a_col_mv, a_row_dup );
      PLA_Copy ( a_col, a_col_dup );
*/
/* end inlined code (commented out) */

/* begin non-inlined code */
      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a_cur, a_col );
      PLA_Obj_view_all ( buf_a_col_dup, & a_col_dup );
      PLA_Obj_view_shift ( a_col_dup, 0, 0, size - nb_alg, 0 );

      PLA_Obj_view_all ( buf_a_row_dup, & a_row_dup );
      PLA_Obj_view_shift ( a_row_dup, 0, 0, 0, size - nb_alg );
      PLA_Obj_set_orientation ( a_col, PLA_PROJ_ONTO_COL );
      PLA_Copy ( a_col, a_row_dup );
      PLA_Copy ( a_col, a_col_dup );
/* end non-inlined code */


/*      pla_syrk_opt_local_update ( min_one, a_col_dup, a_row_dup, one, a_next );*/

      PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
      PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
      PLA_Obj_vert_split_2 ( a_row_dup, 0, & a_row_dup, PLA_DUMMY );

      PLA_Obj_view_all ( a_next, & a_rest );

      while ( TRUE )
      {
         PLA_Obj_split_size ( a_rest, PLA_SIDE_TOP, & size_top, & owner_top );
         PLA_Obj_split_size ( a_rest, PLA_SIDE_LEFT, & size_left, & owner_left );

         if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

         PLA_Obj_view_shift ( a_cur, 0, 0, size, size );
         PLA_Obj_view_shift ( a_col, size, 0, size, 0 );
         PLA_Obj_view_shift ( a_col_dup, size, 0, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, 0, size, 0 );

         PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_TRANS,
                          min_one, a_row_dup, one, a_cur );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          min_one, a_col_dup, a_row_dup, one, a_col );

         PLA_Obj_view_shift ( a_rest, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_cur, size, size, 0, 0 );
         PLA_Obj_view_shift ( a_col, 0, size, 0, 0 );
         PLA_Obj_view_shift ( a_row_dup, 0, size, 0, 0 );
      }
   }

   PLA_Obj_free ( & buf_a_cur_dup );
   PLA_Obj_free ( & buf_a_col_dup );
   PLA_Obj_free ( & buf_a_col_mv );
   PLA_Obj_free ( & buf_a_row_dup );
   PLA_Obj_free ( & a_cur );
   PLA_Obj_free ( & a_cur_dup );
   PLA_Obj_free ( & a_col );
   PLA_Obj_free ( & a_col_dup );
   PLA_Obj_free ( & a_col_mv );
   PLA_Obj_free ( & a_row_dup );
   PLA_Obj_free ( & a_next );
   PLA_Obj_free ( & a_rest );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

/* routine for divide and conquor syrk update */
int pla_syrk_opt_local_update ( PLA_Obj alpha,
                                PLA_Obj dup_col,
                                PLA_Obj dup_row,
                                PLA_Obj beta,
                                PLA_Obj matrix )
{
   int size, top_size, top_owner;
   PLA_Obj top = NULL, left = NULL, right = NULL, bottom = NULL,
           top_left = NULL, bottom_left = NULL, bottom_right = NULL;

   PLA_Obj_global_length ( matrix, & size );
   PLA_Obj_split_size ( matrix, PLA_SIDE_TOP, & top_size, & top_owner );

   if ( size == top_size ) /* everything is local */
   {
      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, alpha, dup_col, beta, matrix );
      return ( PLA_SUCCESS );
   }

   /* matrix distributed across many rows and column */

   size /= 2;

   PLA_Obj_split_4 ( matrix, size, size, & top_left,    PLA_DUMMY,
                                         & bottom_left, & bottom_right );
   PLA_Obj_horz_split_2 ( dup_col, size, & top, & bottom );
   PLA_Obj_vert_split_2 ( dup_row, size, & left, & right );

   PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS, alpha, bottom, left, beta, bottom_left );

   pla_syrk_opt_local_update ( alpha, top, left, beta, top_left );
   pla_syrk_opt_local_update ( alpha, bottom, right, beta, bottom_right );

   PLA_Obj_free ( & top_left );
   PLA_Obj_free ( & bottom_left );
   PLA_Obj_free ( & bottom_right );
   PLA_Obj_free ( & top);
   PLA_Obj_free ( & left );
   PLA_Obj_free ( & right );
   PLA_Obj_free ( & bottom );

   return ( PLA_SUCCESS );
}
