/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/




#include <stdio.h>
#include "mpi.h"
#include "PLA.h"

int   PLA_ALGORYTHMIC_BLOCKSIZE;
 
main(int argc, char *argv[] )
 
/*****************************************************************
PLA_library generic test program
 
Copyright 1996  The University of Texas at Austin
This code is distributed under gnu license.
 
Programmed by Robert van de Geijn
              rvdg@cs.utexas.edu
********************************************************************/
{
  char operation[10],    /* operation to be tested                  */
       side[2],         /* BLAS SIDE parameter                     */
       uplo[2],         /* BLAS UPLO parameter                     */
       transa[2],       /* BLAS TRANS and TRANSA parameter         */
       transb[2],       /* BLAS TRANSB parameter                   */
       diag[2];         /* BLAS DIAG parameter                     */
  int  m,               /* BLAS M parameter                        */
       n,               /* BLAS N parameter                        */ 
       k,               /* BLAS K parameter                        */
       iuseless;        /* resync */
  double alpha,         /* BLAS alpha parameter                    */
       beta,            /* BLAS beta parameter                     */
       duseless;         /* resync rand deal                        */
  int  ia, ja, ib,      /* Scalable BLAS alignment parameters      */
       jb, ic, jc;
  int  nb,alg_nb;       /* Scalable BLAS blocking parameter        */
  int  r, c;            /* logical mesh dimensions                 */
  int  me;              /* my node index                           */
  MPI_Comm comm2d;      /* The 2-D communicator used               */
  int  info=0, nwrong;    /* error information variables             */
  double time, Tmflops, mflops;  /* time and performance variables          */
  PLA_Template template=NULL;
  int  temp_int;
  int memory_alloc_count;
  extern int PLA_MEMORY_ALLOC_COUNT;
#ifdef PNNL
  double SCRATCH_SPACE[100000];
#endif
  char routine_name [ 16 ] = "PLA_rvdg_test";


/* Okay -- since this is SPECIFICALLY to test the memory stuff -- */
/* do that first.                                                 */
#ifndef PNNL
  printf("Look you loopy goof -- you should really have had PNNL\n");
  printf("defined when you built this\n");
#endif

#ifdef PNNL
  PLA_pnnl_malloc_init( (void*)SCRATCH_SPACE, 800000 );
#endif
  

					 /* Initialize MPI                                  */
  MPI_Init( &argc, &argv );
					 /* Extract me                                      */
  MPI_Comm_rank( MPI_COMM_WORLD, &me );

					 /* Get mesh info and broadcast                     */
  if ( me == 0 )
  {
	 printf("Input #procs/row, #procs/col\n");
	 scanf( "%d%d", &r, &c );
  }
  MPI_Bcast( &r, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &c, 1, MPI_INT, 0, MPI_COMM_WORLD );
					 /* Get blocking parameter, nb, and broadcast       */
  if ( me == 0 )
  {
	 printf("Input nb alg_nb\n");
	 scanf( "%d%d", &nb, &alg_nb );
  }
  MPI_Bcast( &nb, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &alg_nb, 1, MPI_INT, 0, MPI_COMM_WORLD );
  if (0==alg_nb) 
     PLA_ALGORYTHMIC_BLOCKSIZE=nb;
  else
     PLA_ALGORYTHMIC_BLOCKSIZE=alg_nb;   
  PLA_Comm_1D_to_2D (MPI_COMM_WORLD,
                     r,
                     c,
                     &comm2d);

  PLA_Init (comm2d);
  PLA_Stack_push_current ( routine_name );
  PLA_Temp_create(nb, 0, &template );

  if (0!=alg_nb) pla_Environ_set_nb_alg (PLA_OP_ALL_ALG, alg_nb);

					 /* Get alignment info and bcast                    */
  if ( me == 0 ) {
/* weird -- if this was taken from sB_test -- there were only 6 */
/* strip this comment out                                       */
	 printf("Input ia, ja, ib, jb,  ic, jc\n");
	 scanf( "%d%d%d%d%d%d", &ia, &ja, &ib, &jb, &ic, &jc );
  }
  MPI_Bcast( &ia, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &ja, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &ib, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &jb, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &ic, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &jc, 1, MPI_INT, 0, MPI_COMM_WORLD );
					 /* Get test problems and execute                   */
  for (;;) {
					 /* get operation and BLAS parameters and bcast     */
	 if ( me == 0 ){
		/* printf("Input operation side uplo transa transb diag m n k\n"); */
		scanf( "%s%s%s%s%s%s%d%d%d",
				 operation, side, uplo, transa, transb, diag,
				 &m, &n, &k );
		alpha = 1.0; beta = 0.0;
					 /* for some reason, the paragon is not reading %f */
	 /* Actually, in our scheme it would probably be better to     */
	 /* create the alpha and beta Mscalar objects here             */
	 }
	 MPI_Bcast( operation, 10, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( side, 2, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( uplo, 2, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( transa, 2, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( transb, 2, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( diag, 2, MPI_CHAR, 0, MPI_COMM_WORLD );
	 MPI_Bcast( &m, 1, MPI_INT, 0, MPI_COMM_WORLD );
	 MPI_Bcast( &n, 1, MPI_INT, 0, MPI_COMM_WORLD );
	 MPI_Bcast( &k, 1, MPI_INT, 0, MPI_COMM_WORLD );
	 MPI_Bcast( &alpha, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
	 MPI_Bcast( &beta,  1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
					 /* call appropriate test routine                   */

         memory_alloc_count = PLA_MEMORY_ALLOC_COUNT;
         if ( operation[0] == '#' )
         {
            time=0.0e+0;
            mflops=0.0e+0;
            info=0;
         }
         else if ( strcmp( operation, "ZGEMM" ) == 0 || 
                   strcmp( operation, "ZGEMM1") == 0 || 
                   strcmp( operation, "ZGEMM2") == 0 || 
                   strcmp( operation, "ZGEMM3") == 0 )
	 {
                if ( !strcmp( operation, "ZGEMM" ) ) temp_int = 0;
                if ( !strcmp( operation, "ZGEMM1") ) temp_int = 1;
                if ( !strcmp( operation, "ZGEMM2") ) temp_int = 2;
                if ( !strcmp( operation, "ZGEMM3") ) temp_int = 3;

		if( ((m*k) <= 160000) && ((n*k) <= 160000) && ((m*n)<= 160000))
		{
		PLA_test_PLA_zgemm( temp_int, 
				  template, transa, transb, m, n, k, alpha, beta,
						ia, ja, ib, jb, ic, jc,
						&time, &mflops, &info );
		}
                else
		{
		PLA_no_test_PLA_zgemm( temp_int, 
				      template, transa, transb, m, n, k, alpha, beta,
						ia, ja, ib, jb, ic, jc,
						&time, &mflops, &info );
		iuseless = 42;
		duseless = PLA_random(iuseless);
		}
	 }
         else if ( strcmp( operation, "DGEMM" ) == 0 || 
                   strcmp( operation, "DGEMM1") == 0 || 
                   strcmp( operation, "DGEMM2") == 0 || 
                   strcmp( operation, "DGEMM3") == 0 )
	 {
                if ( !strcmp( operation, "DGEMM" ) ) temp_int = 0;
                if ( !strcmp( operation, "DGEMM1") ) temp_int = 1;
                if ( !strcmp( operation, "DGEMM2") ) temp_int = 2;
                if ( !strcmp( operation, "DGEMM3") ) temp_int = 3;

		if( ((m*k) <= 160000) && ((n*k) <= 160000) && ((m*n)<= 160000))
		{
		PLA_test_PLA_dgemm( temp_int, 
				  template, transa, transb, m, n, k, alpha, beta,
						ia, ja, ib, jb, ic, jc,
						&time, &mflops, &info );
		}
                else
		{
		PLA_no_test_PLA_dgemm( temp_int, 
				      template, transa, transb, m, n, k, alpha, beta,
						ia, ja, ib, jb, ic, jc,
						&time, &mflops, &info );
		iuseless = 42;
		duseless = PLA_random(iuseless);
		}
	 }
         else if ( strcmp( operation, "DSYRK" ) == 0 )
	 {
                alpha=beta=1.0;
		if( ((n*k) <= 160000) && ((n*n) <= 160000)) 
		   PLA_test_dsyrk( template, uplo, transa, n, k, alpha, beta,
						ia, ja, ic, jc,
						&time, &mflops, &info );
		else
		  PLA_no_test_dsyrk( template, uplo, transa, n, k, alpha, beta,
						ia, ja, ic, jc,
                                                &time, &mflops, &info ); 
		iuseless = 42;
		duseless = PLA_random(iuseless);
         } 
         else if ( strcmp( operation, "DSYMM" ) == 0 )
	 {
                alpha=beta=1.0;
		if( ((n*m) <= 160000) &&((n*n) <= 160000) && ((m*m) <= 160000)) 
		   PLA_test_dsymm( template, side, uplo, m, n, alpha, beta,
						ia, ja, ib, jb, ic, jc,
						&time, &mflops, &info );
/*		else
		  PLA_no_test_dsyrk( template, uplo, transa, n, k, alpha, beta,
						ia, ja, ic, jc,
                                                &time, &mflops, &info ); 
		iuseless = 42;
		duseless = PLA_random(iuseless); */
         } 
         else if ( strcmp( operation, "DCHOL" ) == 0 )
	 {
		if( ((n*n) <= 160000)) 
                   PLA_test_dchol ( template, uplo, n, ia, ja, &time, &mflops, &info );
		else
		   PLA_no_test_dchol( template, uplo, n, ia, ja, &time, &mflops, &info );
		iuseless = 42;
		duseless = PLA_random(iuseless);
         }      
         else if ( strcmp( operation, "CHOLxL1" ) == 0 ||
                   strcmp( operation, "CHOLxL2" ) == 0 ||
                   strcmp( operation, "CHOLxL3" ) == 0 ||
                   strcmp( operation, "CHOLxL4" ) == 0 ||
                   strcmp( operation, "CHOLxL5" ) == 0 ||
                   strcmp( operation, "CHOLxL6" ) == 0 ||
                   strcmp( operation, "CHOLxR1" ) == 0 ||
                   strcmp( operation, "CHOLxR2" ) == 0 ||
                   strcmp( operation, "CHOLxR3" ) == 0 ||
                   strcmp( operation, "CHOLxR4" ) == 0 ||
                   strcmp( operation, "CHOLxR5" ) == 0 ||
                   strcmp( operation, "CHOLxR6" ) == 0 ||
                   strcmp( operation, "CHOLxR7" ) == 0 )
	 {
                if ( !strcmp( operation, "CHOLxL1") ) temp_int = 1;
                if ( !strcmp( operation, "CHOLxL2") ) temp_int = 2;
                if ( !strcmp( operation, "CHOLxL3") ) temp_int = 3;
                if ( !strcmp( operation, "CHOLxL4") ) temp_int = 4;
                if ( !strcmp( operation, "CHOLxL5") ) temp_int = 5;
                if ( !strcmp( operation, "CHOLxL6") ) temp_int = 6;
                if ( !strcmp( operation, "CHOLxR1") ) temp_int = 11;
                if ( !strcmp( operation, "CHOLxR2") ) temp_int = 12;
                if ( !strcmp( operation, "CHOLxR3") ) temp_int = 13;
                if ( !strcmp( operation, "CHOLxR4") ) temp_int = 14;
                if ( !strcmp( operation, "CHOLxR5") ) temp_int = 15;
                if ( !strcmp( operation, "CHOLxR6") ) temp_int = 16;
                if ( !strcmp( operation, "CHOLxR7") ) temp_int = 17;

		if( ((n*n) <= 160000)) 
                   PLA_test_opt_dchol ( temp_int, template,
                                            uplo, n, ia, ja, &time, &mflops, &info );
		else
                   PLA_no_test_opt_dchol ( temp_int, template,
                                               uplo, n, ia, ja, &time, &mflops, &info );
		iuseless = 42;
		duseless = PLA_random(iuseless);
         }      
 
         else if ( strcmp( operation, "DGEMV" ) == 0 )
	 {
		PLA_test_dgemv ( template, uplo, transa, m, n, alpha, beta,
					ia, ja, ic, jc,
                                          &time, &mflops, &info ); 
		iuseless = 42;
		duseless = PLA_random(iuseless);
         }      
    else if ( strcmp( operation, "DSYR2K" ) == 0 )
    {
      if( ((n*k) <= 160000) && ((n*n) <= 160000)) {
      PLA_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info ); }
     /* else
      {
      PLA_no_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      } */ 
    }
    else if ( strcmp( operation, "DSYMV" ) == 0 )
    {
      if( ((n*n) <= 160000) && ((n*n) <= 160000)) {
      PLA_test_dsymv( template, uplo, n, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info ); }
     /* else
      {
      PLA_no_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      } */ 
    }
    else if ( strcmp( operation, "DTRMV" ) == 0 )
    {
      if( ((n*n) <= 160000) && ((n*n) <= 160000)) {
      PLA_test_dtrmv( template, uplo, transa, diag,  n,
                  ia, ja, ib, jb, &time, &mflops, &info ); }
     /* else
      {
      PLA_no_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      } */ 
    }
    else if ( strcmp( operation, "DSYR" ) == 0 )
    {
      if( ((n*n) <= 160000) && ((n*n) <= 160000)) {
      PLA_test_dsyr( template, uplo, n, alpha,
                  ia, ja, ib, jb, &time, &mflops, &info ); }
     /* else
      {
      PLA_no_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      } */ 
    }
    else if ( strcmp( operation, "DSYR2" ) == 0 )
    {
      if( ((n*n) <= 160000) && ((n*n) <= 160000)) {
      PLA_test_dsyr2( template, uplo, n, alpha,
                  ia, ja, ib, jb, ic, jc, &time, &mflops, &info ); }
     /* else
      {
      PLA_no_test_PLA_dsyr2k( template, uplo, transa, n, k, alpha, beta, 
                  ia, ja, ib, jb, ic, jc,
                  &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      } */ 
    }
    else if ( strcmp( operation, "DTRSM" ) == 0 )
    {
      if( ((m*m) <= 160000) && ((m*n)<= 160000))
      {
      PLA_test_dtrsm( template, side, uplo, transa, diag, m, n,  alpha, 
                  ia, ja, ib, jb, &time, &mflops, &info );
      }
     /* else
      {
      PLA_test_PLA_dtrsm( template, side, uplo, transa, diag, m, n, alpha, 
                  ia, ja, ib, jb, &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      }*/
    } 
    else if ( strcmp( operation, "DTRMM" ) == 0 )
    {
      if( ((m*m) <= 160000) && ((m*n)<= 160000))
      {
      PLA_test_dtrmm( template, side, uplo, transa, diag, m, n,  alpha, 
                  ia, ja, ib, jb, &time, &mflops, &info );
      }
     /* else
      {
      PLA_test_PLA_dtrsm( template, side, uplo, transa, diag, m, n, alpha, 
                  ia, ja, ib, jb, &time, &mflops, &info );
      iuseless = 42;
      duseless = PLA_random(iuseless);
      }*/
    } 
    else if ( strcmp( operation, "DONE" ) == 0 )
      break;
    else if ( me == 0 ) {
      printf( "%s operation not found\n", operation );
      continue;
    }
                /* get number of wrong entries on all nodes        */
    MPI_Allreduce( &info, &nwrong, 1, MPI_INT, MPI_SUM, 
                   MPI_COMM_WORLD );
                /* print results                                   */
    MPI_Barrier(MPI_COMM_WORLD);
    if ( me == 0 ) {
      Tmflops = mflops * r * c;
/*      printf( "%5d %5d %5d     %5d %5d       %10.4lf       %10.4lf\n", m, n, k, nb, alg_nb, mflops, Tmflops); */
      printf ( "%8s %c %c %c %c %c %5d %5d %5d %5d %5d %10.4lf %10.4lf\n",
                operation, side[0], uplo[0], transa[0], transb[0], diag[0],
                m, n, k, nb, alg_nb, mflops, Tmflops );
      fflush(stdout);
    };
    if ( ( memory_alloc_count -= PLA_MEMORY_ALLOC_COUNT ) != 0 )
    {
       printf ( "PE%d : WARNING! Memory allocation discrepency of %d after %6s\n",
                me, -memory_alloc_count, operation );
       /*exit ( 0 );*/
    }
    if ( nwrong != 0 ) exit( 0 );
  };
                /* clean up MPI                                    */
  PLA_Temp_free(&template);
  PLA_Stack_pop_current ( routine_name );
  PLA_Finalize ();
  MPI_Finalize();
}
 
