/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/

#if MANUFACTURE == CRAY
#define    PLA_OP_PAN_PAN_DEFAULT       128
#define    PLA_OP_PAN_MAT_DEFAULT       128
#define    PLA_OP_MAT_PAN_DEFAULT       128
#define    PLA_OP_PAN_TRIAN_MAT_DEFAULT 128 
#define    PLA_OP_SYM_PAN_PAN_DEFAULT   128 
#define    PLA_OP_TRIAN_MAT_PAN_DEFAULT 128 
 
#elif MANUFACTURE == INTEL
#define    PLA_OP_PAN_PAN_DEFAULT        20
#define    PLA_OP_PAN_MAT_DEFAULT        20
#define    PLA_OP_MAT_PAN_DEFAULT        20
#define    PLA_OP_PAN_TRIAN_MAT_DEFAULT  20 
#define    PLA_OP_SYM_PAN_PAN_DEFAULT    20 
#define    PLA_OP_TRIAN_MAT_PAN_DEFAULT  20 

#elif MANUFACTURE == IBM
#define    PLA_OP_PAN_PAN_DEFAULT       100
#define    PLA_OP_PAN_MAT_DEFAULT       100
#define    PLA_OP_MAT_PAN_DEFAULT       100
#define    PLA_OP_PAN_TRIAN_MAT_DEFAULT 100 
#define    PLA_OP_SYM_PAN_PAN_DEFAULT   100 
#define    PLA_OP_TRIAN_MAT_PAN_DEFAULT 100 

#else
#define    PLA_OP_PAN_PAN_DEFAULT        24
#define    PLA_OP_PAN_MAT_DEFAULT        24
#define    PLA_OP_MAT_PAN_DEFAULT        24
#define    PLA_OP_PAN_TRIAN_MAT_DEFAULT  24 
#define    PLA_OP_SYM_PAN_PAN_DEFAULT    24 
#define    PLA_OP_TRIAN_MAT_PAN_DEFAULT  24 
#endif



static int 
  pan_pan       = PLA_OP_PAN_PAN_DEFAULT,
  pan_mat       = PLA_OP_PAN_MAT_DEFAULT,
  mat_pan       = PLA_OP_MAT_PAN_DEFAULT,
  pan_trian_mat = PLA_OP_PAN_TRIAN_MAT_DEFAULT,
  sym_pan_pan   = PLA_OP_SYM_PAN_PAN_DEFAULT,
  trian_mat_pan = PLA_OP_TRIAN_MAT_PAN_DEFAULT;



#include "PLA.h"

int PLA_Environ_nb_alg (int algorythm_type, 
			PLA_Template  template,
			int *algorythm_blocksize)
{
  int error=0;

  switch (algorythm_type) {
  case PLA_OP_PAN_PAN:
    (*algorythm_blocksize) = pan_pan;
    break;
  case PLA_OP_PAN_MAT:
    (*algorythm_blocksize) = pan_mat;
    break;
  case PLA_OP_MAT_PAN:
    (*algorythm_blocksize) = mat_pan;
    break;
  case PLA_OP_PAN_TRIAN_MAT:
    (*algorythm_blocksize) = pan_trian_mat;
    break;
  case PLA_OP_SYM_PAN_PAN:
    (*algorythm_blocksize) = sym_pan_pan;
    break;
  case PLA_OP_TRIAN_MAT_PAN:
    (*algorythm_blocksize) = trian_mat_pan;
    break;
  default:
    printf ("Invalid algorythm code passed to PLA_Environ_nb_alg\n");
    printf ("Code %d\n", algorythm_type);
    error = 1;
  }
  return (error);
}


int pla_Environ_set_nb_alg (int algorythm_type,  
		            int algorythm_blocksize)
{
  int error=0;

  if (0>algorythm_blocksize) {
    printf ("Invalid algorythm blocksize passed to pla_Environ_set_nb_alg\n");
    printf ("blocksize %d must be >=0 \n", algorythm_blocksize);
    error = 1;
  }    
  switch (algorythm_type) {
  case PLA_OP_PAN_PAN:
    if (0==algorythm_blocksize)
      pan_pan = PLA_OP_PAN_PAN_DEFAULT;
    else
      pan_pan = algorythm_blocksize;
    break;
  case PLA_OP_PAN_MAT:
    if (0==algorythm_blocksize)
      pan_mat = PLA_OP_PAN_MAT_DEFAULT;
    else
      pan_mat = algorythm_blocksize;
    break;
  case PLA_OP_MAT_PAN:
    if (0==algorythm_blocksize)
      mat_pan = PLA_OP_MAT_PAN_DEFAULT;
    else
      mat_pan = algorythm_blocksize;
    break;
  case PLA_OP_PAN_TRIAN_MAT:
    if (0==algorythm_blocksize)
      pan_trian_mat = PLA_OP_PAN_TRIAN_MAT_DEFAULT;
    else
      pan_trian_mat = algorythm_blocksize;
    break;
  case PLA_OP_SYM_PAN_PAN:
    if (0==algorythm_blocksize)
      sym_pan_pan = PLA_OP_SYM_PAN_PAN_DEFAULT;
    else
      sym_pan_pan = algorythm_blocksize;
    break;
  case PLA_OP_TRIAN_MAT_PAN:
    if (0==algorythm_blocksize)
      trian_mat_pan = PLA_OP_TRIAN_MAT_PAN_DEFAULT;
    else
      trian_mat_pan = algorythm_blocksize;
    break;
  case PLA_OP_ALL_ALG:
    if (0==algorythm_blocksize) {
      pan_pan = PLA_OP_PAN_PAN_DEFAULT;
      pan_mat = PLA_OP_PAN_MAT_DEFAULT;
      mat_pan = PLA_OP_MAT_PAN_DEFAULT;
      pan_trian_mat = PLA_OP_PAN_TRIAN_MAT_DEFAULT;
      sym_pan_pan   = PLA_OP_SYM_PAN_PAN_DEFAULT;
      trian_mat_pan = PLA_OP_TRIAN_MAT_PAN_DEFAULT;
    }
    else {
      pan_pan = algorythm_blocksize;
      pan_mat = algorythm_blocksize;
      mat_pan = algorythm_blocksize;
      pan_trian_mat = algorythm_blocksize;
      sym_pan_pan   = algorythm_blocksize;
      trian_mat_pan = algorythm_blocksize;
    }
    break;
  default:
    printf ("Invalid algorythm code passed to pla_Environ_set_nb_alg\n");
    printf ("Code %d\n", algorythm_type);
    error = 1;
  }
  return (error);
}

