
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/


#include "PLA.h"
#include "PLA_Fill_laobj.h"

int PLA_Fill_laobj( PLA_Obj laobj)
/******************************************************
  Fills in a matrix with numbers of the form i + 0.01*j
  for the (local) i,j th entry in the matrix.  
*******************************************************/
{
  int 
    length,
    width,
    i, j,
    lld;
  
  void *void_ptr;
  MPI_Datatype datatype;
  
  PLA_Obj_datatype( laobj, &datatype);

  if( MPI_DOUBLE != datatype && MPI_FLOAT != datatype)
    {
      printf("Illegal datatype passed to fill_dbl_matrix()\n");
      return 2;
    }
      
  PLA_Obj_local_buffer( laobj, &void_ptr);
  PLA_Obj_local_length( laobj, &length);
  PLA_Obj_local_width( laobj, &width);
  PLA_Obj_local_ldim( laobj, &lld);
  
  
  for( j=0; j< width; j++)
    {
      for( i=0; i<length; i++)
	{
	  if( MPI_DOUBLE == datatype)
	    {
	      ((double *)void_ptr)[ j*lld + i] = 0.01*j + i;
	    }
	  if( MPI_FLOAT == datatype)
	    {
	      ((float *)void_ptr)[ j*lld + i] = 0.01*j + i;
	    }
	}
    }
  
  return 0;
}



int PLA_Print_laobj( PLA_Obj laobj)
/*********************************************************
  Prints out the local data in an PLA_Obj.  Note that the
  datatype MUST be MPI_DOUBLE.
*********************************************************/
{
  int 
    length,
    width,
    i, j,
    lld;
  
  double *buffer;
  void *void_ptr;
  MPI_Datatype datatype;
  
  PLA_Obj_datatype( laobj, &datatype);

  PLA_Obj_local_buffer( laobj, &void_ptr);
  PLA_Obj_local_length( laobj, &length);
  PLA_Obj_local_width( laobj, &width);
  PLA_Obj_local_ldim( laobj, &lld);
  
  for( i=0; i< length; i++)
    {
      for( j=0; j<width; j++)
	{
	  if( MPI_DOUBLE == datatype)
	    {
	      printf("%8.4e  ", ((double *)void_ptr)[ j*lld + i]);
	    }
	  else if( MPI_FLOAT == datatype)
	    {
	      printf("%lf  ", ((float *)void_ptr)[ j*lld + i]);
	    }
	  else if ( MPI_DOUBLE_COMPLEX == datatype)
	    {
	      printf(" %lf + ", ((double *)void_ptr)[ j*lld*2 + i*2]);
	      printf(" %lf i  ", ((double *)void_ptr)[ j*lld*2 + i*2+1]);
	    }
	  else if( MPI_COMPLEX == datatype)
	    {
	      printf("%lf  ", ((float *)void_ptr)[ j*lld*2 + i]);
	      printf("%lf  ", ((float *)void_ptr)[ j*lld*2 + i+1]);
	    }
	if (j%6==0 && j>0) printf("\n");

	}
      printf("\n");
    }
  
  return 0;
}



