/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu

***************************************************************************/



#include "PLA.h"
#include "PLA_API.h"

int PLA_Obj_set_to_zero( PLA_Obj laobj) 
{
int  zero=0;
PLA_Obj_set( laobj, MPI_INT, &zero);
}

int PLA_Obj_set_to_one ( PLA_Obj laobj) 
{
int one=1;
PLA_Obj_set( laobj, MPI_INT, &one);
}

int PLA_Obj_set_to_minus_one ( PLA_Obj laobj) 
{
int mone=-1;
PLA_Obj_set( laobj, MPI_INT, &mone);
}




int PLA_Obj_set( PLA_Obj laobj, 
                 MPI_Datatype   datatype, 
                 void *value_ptr)
/*****************************************************************

  Copies value into the buffer of an PLA_Obj.  The address of the data
  is passed in as a void pointer.  
  
*****************************************************************/
{
  int  
	 local_length,
	 local_width,
	 col,
	 stride,
	 lld,
	 i_zero=0;

  int    *int_value_ptr;
  float  *float_value_ptr;
  double *double_value_ptr;
  int    int_value;
  float  float_value;
  double double_value;
  float  complex_float[2]; /* Complex cases */
  double complex_double[2];

  MPI_Datatype obj_datatype;
  int type_size;

  void *buffer;

  PLA_Obj_obj_type obj_type;

  if (MPI_INT == datatype)
	 int_value_ptr = (int *)(value_ptr);
  else if ((MPI_FLOAT == datatype) || (MPI_COMPLEX == datatype))
	 float_value_ptr = (float *)(value_ptr);
  else if ((MPI_DOUBLE == datatype) || (MPI_DOUBLE_COMPLEX == datatype))
	 double_value_ptr = (double *)(value_ptr);


  PLA_Obj_local_all( laobj,
			  &obj_datatype,
			  &local_length,
			  &local_width,
			  &buffer,
			  &lld);

  PLA_Obj_objtype( laobj, &obj_type);

  if( PLA_VECTOR == obj_type)
	 {
		PLA_Obj_local_stride( laobj, &stride);
	 }
  else
	 stride = 1;


  MPI_Type_size (obj_datatype, &type_size);

  for( col = 0; col < local_width; col++)
	 {
		if( MPI_DOUBLE == obj_datatype || sizeof(double) ==  type_size )
	{
			 if (MPI_INT == datatype)
				double_value = *int_value_ptr;
			 else if (MPI_FLOAT == datatype)
				double_value = *float_value_ptr;
			 else
				double_value = *double_value_ptr;

			 PLA_dcopy( &local_length,
							&double_value,
							&i_zero,
							&((double *)buffer)[ col*lld],
							&stride);
	}
		else if( MPI_FLOAT == obj_datatype || sizeof(float) ==  type_size )
	{
			 if (MPI_INT == datatype)
				float_value = *int_value_ptr;
			 else if (MPI_FLOAT == datatype)
				float_value = *float_value_ptr;
			 else
				float_value = *double_value_ptr;

	  PLA_scopy( &local_length,
			&float_value,
			&i_zero,
			&((float *)buffer)[ col*lld],
			&stride);
	}
	else if( MPI_COMPLEX == obj_datatype || 2*sizeof(float) ==  type_size )
	{
		 if (MPI_INT == datatype)
		 {
			complex_float[0] = *int_value_ptr;
			complex_float[1] = 0.0;
		 }
		 else if (MPI_FLOAT == datatype)
		 {
			complex_float[0] = *float_value_ptr;
			complex_float[1] = 0.0;
		 }
		 else if (MPI_FLOAT == datatype)
		 {
			complex_float[0] = *double_value_ptr;
			complex_float[1] = 0.0;
		 }
		 else if (MPI_COMPLEX == datatype)
		 {
			complex_float[0] = *float_value_ptr;
			complex_float[1] = *(float_value_ptr+1);
		 }
		 else if (MPI_DOUBLE_COMPLEX == datatype)
		 {
			complex_float[0] = *double_value_ptr;
			complex_float[1] = *(double_value_ptr+1);
		 }
		 PLA_ccopy( &local_length,
		 &complex_float,
		 &i_zero,
		 &((float *)buffer)[ 2*col*lld],
		 &stride);
	}
	else if( MPI_DOUBLE_COMPLEX == obj_datatype || 
                 2*sizeof(double) ==  type_size )
	{
		 if (MPI_INT == datatype)
		 {
			complex_double[0] = *int_value_ptr;
			complex_double[1] = 0.0;
		 }
		 else if (MPI_FLOAT == datatype)
		 {
			complex_double[0] = *float_value_ptr;
			complex_double[1] = 0.0;
		 }
		 else if (MPI_FLOAT == datatype)
		 {
			complex_double[0] = *double_value_ptr;
			complex_double[1] = 0.0;
		 }
		 else if (MPI_COMPLEX == datatype)
		 {
			complex_double[0] = *float_value_ptr;
			complex_double[1] = *(float_value_ptr+1);
		 }
		 else if (MPI_DOUBLE_COMPLEX == datatype)
		 {
			complex_double[0] = *double_value_ptr;
			complex_double[1] = *(double_value_ptr+1);
		 }
		 PLA_zcopy( &local_length,
		 complex_double,
		 &i_zero,
		 &((double *)buffer)[ 2*col*lld],
		 &stride);
	 }
  }
  return 0;
}

int PLA_Obj_set_to_identity ( PLA_Obj laobj ) 
{
   int i, state, length, width;
   MPI_Datatype datatype;
   int int_one = 1;
   float s_one = 1.0e+0;
   double d_one = 1.0e+0;
   void * value;

   PLA_Obj_global_length ( laobj, & length );
   PLA_Obj_global_width ( laobj, & width );
   PLA_Obj_datatype ( laobj, & datatype );

   if ( datatype == MPI_INT ) value = & int_one;
   else if ( datatype == MPI_FLOAT ) value = & s_one;
   else if ( datatype == MPI_DOUBLE ) value = & d_one;


   PLA_Obj_set_to_zero ( laobj );

   PLA_API_state ( & state );
   if ( state == PLA_STATE_NORMAL ) PLA_API_begin ();

   PLA_Obj_API_open ( laobj );

   for ( i = 0 ; i < min ( length, width ) ; i++ )
      PLA_API_axpy_matrix_to_global ( 1, 1, value, value, 1, laobj, i, i );

   PLA_Obj_API_close ( laobj );
   if ( state == PLA_STATE_NORMAL ) PLA_API_end ();

   return ( PLA_SUCCESS );

}
