
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/

#include "PLA.h"

int PLA_Extract_matrix( PLA_Obj global_matrix, PLA_Obj local_matrix)
/******************************************************************
  Extracts the local portion of a global matrix.  The global matrix
  should be contained in a multiscalar duplicated on each calling 
  node.
*******************************************************************/
{
  PLA_Obj
    current_local_matrix    = NULL,
    current_local_panel     = NULL,
    remainder_local_matrix  = NULL,
    current_local_block     = NULL,
    remainder_local_panel   = NULL,
    
    current_global_matrix   = NULL,
    current_global_panel    = NULL,
    remainder_global_matrix = NULL,
    current_global_block    = NULL,
    remainder_global_panel  = NULL;

  int 
    myrow, 
    mycol,
    left_size,
    top_size,
    current_proc_col,
    current_proc_row,
    idummy,
    obj_type,
    error,
    local_leading_dimension;

  void *current_local_block_buffer;
  
  
  PLA_Template template;
  
  PLA_Obj_template( local_matrix, &template);
  PLA_Temp_mycoords( template, &myrow, &mycol);

  PLA_Obj_local_ldim( local_matrix, &local_leading_dimension);

  PLA_Matrix_submatrix( local_matrix, PLA_DIM_ALL, PLA_DIM_ALL,
		      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_local_matrix);

  PLA_Mscalar_submscalar( global_matrix, PLA_DIM_ALL, PLA_DIM_ALL,
		      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_global_matrix);
  
  for(;;)
    {
      /* find the leftmost contiguous block of columns that's */
      /* owned by one col of proc */
      PLA_Obj_left( current_local_matrix, &left_size, &current_proc_col);
      
      /* exit if no more columns left */
      if( left_size == 0) break;
      
      /* split the matrices according to the previous call */
      error = PLA_Obj_vert_split_2( current_local_matrix, left_size, 
			    &current_local_panel, &remainder_local_matrix);

      PLA_Obj_vert_split_2( current_global_matrix, left_size, 
			    &current_global_panel, &remainder_global_matrix);
      
      
      /* now process the current panel */
      for(;;)
	{
	  /* find the topmost contiguous block of rows owned by one proc. */
	  PLA_Obj_top( current_local_panel, &top_size, &current_proc_row);

	  /* exit this loop if no rows left */
	  if( top_size == 0) break;
	  
	  /* split the panels according to the previous call */
	  error = PLA_Obj_horz_split_2( current_local_panel, top_size, 
				&current_local_block, &remainder_local_panel);

	  PLA_Obj_horz_split_2( current_global_panel, top_size, 
				&current_global_block, &remainder_global_panel);
      
	  /* copy the contents from the global block to the local block */
	  if( myrow == current_proc_row && mycol == current_proc_col)
	    {
	      /* first get the buffer of the current local block */
	      PLA_Obj_local_buffer( current_local_block, &current_local_block_buffer);
	      
	      /* and extract the contents of the global block there */
	      PLA_Obj_extract_contents( current_global_block, 
					local_leading_dimension,
					1,
					&idummy, &idummy,
					current_local_block_buffer);
	    }
	  
	  /* update the panel views */
	  PLA_Obj_swap( &current_local_panel, &remainder_local_panel);
	  PLA_Obj_swap( &current_global_panel, &remainder_global_panel);

	}  /* end of the panel for loop */
      
      /* update the matrix views */
      PLA_Obj_swap( &current_local_matrix, &remainder_local_matrix);
      PLA_Obj_swap( &current_global_matrix, &remainder_global_matrix);

    }  /* end of matrix for loop */

  /* free all the views */
  PLA_Obj_free( &current_local_matrix);
  PLA_Obj_free( &current_global_matrix);

  PLA_Obj_free( &current_local_panel);
  PLA_Obj_free( &current_global_panel);

  PLA_Obj_free( &remainder_local_matrix);
  PLA_Obj_free( &remainder_global_matrix);

  PLA_Obj_free( &remainder_local_panel);
  PLA_Obj_free( &remainder_global_panel);
  

  return 0;
  
} /* END of PLA_Extract_matrix() */

      
      
	  
	      
