/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu

***************************************************************************/
#include "PLA.h"

int PLA_Obj_set_contents( PLA_Obj laobj, void *value_ptr)
/*****************************************************************

  Copies data into the buffer of an PLA_Obj.  The address of the data
  is passed in as a void pointer.  The source data is accessed with
  stride 1, regardless of the stride of the LA_Obj (for type PLA_VECTOR).

*****************************************************************/
{
  int
	 local_length,
	 local_width,
	 col,
	 stride,
	 lld,
	 i_one=1;

  int type_size;

  void *buffer;

  MPI_Datatype datatype;

  PLA_Obj_obj_type obj_type;

  PLA_Obj_local_all( laobj,
			  &datatype,
			  &local_length,
			  &local_width,
			  &buffer,
			  &lld);

  PLA_Obj_objtype( laobj, &obj_type);

  if( PLA_VECTOR == obj_type)
	 {
		PLA_Obj_local_stride( laobj, &stride);
	 }
  else
	 stride = 1;


  MPI_Type_size (datatype, &type_size);

  for( col = 0; col < local_width; col++)
  {
	if( MPI_DOUBLE == datatype || sizeof(double) ==  type_size )
	{
	  PLA_dcopy( &local_length,
		 &((double *)value_ptr)[ col*local_length],
		 &i_one,
		 &((double *)buffer)[ col*lld],
		 &stride);
	}
	else if( MPI_FLOAT == datatype || sizeof(float) ==  type_size )
	{
	  PLA_scopy( &local_length,
		 &((float *)value_ptr)[ col*local_length],
		 &i_one,
		 &((float *)buffer)[ col*lld],
		 &stride);
	}
	else if( MPI_COMPLEX == datatype || 2*sizeof(float) ==  type_size )
	{
	  PLA_ccopy( &local_length,
		 &((float *)value_ptr)[ 2*col*local_length],
		 &i_one,
		 &((float *)buffer)[ 2*col*lld],
		 &stride);
	}
	else if( MPI_DOUBLE_COMPLEX == datatype || 2*sizeof(double) ==  type_size )
	{
	  PLA_zcopy( &local_length,
		 &((double *)value_ptr)[ 2*col*local_length],
		 &i_one,
		 &((double *)buffer)[ 2*col*lld],
		 &stride);
	}
  }
  return 0;
}



			
		 
  
