#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_count_at(PLA_Obj X, int* n_0, int* n_half, int* n_1 ) {
/*
    Determine the number of eigenvalues of X at 0, 1/2, and 1.  
*/
    int            global_length;
    double         trX, trX2, r_0, r_half, r_1;
    void           *trace_buffer;
    PLA_Obj        X2=NULL, trace=NULL;
    MPI_Datatype   datatype;
    PLA_Template   template;

    /* Extract info from matrix X */
    PLA_Obj_datatype       ( X, &datatype );
    PLA_Obj_global_length  ( X, &global_length );
    PLA_Obj_template       ( X, &template );

    /* Temporary matrix X2 = X*X */
    PLA_Matrix_create_conf_to( X, &X2 ); 
    PLA_Obj_set_to_zero( X2 );
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, X, X, PRISM_zero, X2 );

    /* Temporary objects for trace operation */
    PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
        template, &trace );
    PLA_Obj_local_buffer( trace, &trace_buffer );

    /* Compute trace(X) */
    PRISM_trace( X, trace );
    trX = *(double *)trace_buffer;

    /* Compute trace(X2) */
    PRISM_trace( X2, trace );
    trX2 = *(double *)trace_buffer;

    r_0    = ABS( global_length + 2.*trX2 - 3.*trX );
    r_half = 4. * ABS( trX2 - trX );
    r_1    = ABS( 2.*trX2 - trX );

    /* Round results to integer values */
    *n_0    = floor( r_0 + .5 );
    *n_half = floor( r_half + .5 );
    *n_1    = floor( r_1 + .5 );

    /* Haven't decided what do do in this case */
    if ( *n_0 + *n_half + *n_1 != global_length ) 
       printf("Error in PRISM_count_at\n");

    PLA_Obj_free( &X2 );
    PLA_Obj_free( &trace );

#ifdef DB_GLOBAL_
    if (me==ROOT) {
       printf(">> PRISM_count_at\n");
       printf("n_0=%d  n_half=%d  n_1=%d\n", *n_0, *n_half, *n_1);
    }
#endif
}

