#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_syisda(PLA_Obj A, PLA_Obj Z, PLA_Obj D, int cut_off)
{
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj cur_est = NULL;
  PRISM_final_list final_list;
  double t_start;

  /* This should be made into a formal error check */
  cut_off = MAX(cut_off,2);

  /* Get information about matrix A */
  PLA_Obj_template(A, &template);
  PLA_Obj_datatype(A, &datatype);

  /* Initialize global objects */
  PRISM_init(template);

  /* Multiscalar to store lower bound of max evalue of A */
  PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
    template, &cur_est);

  /* Initialize Z - this is also done in find_subpspaces */
  PRISM_set_to_ident(Z);

  /* Initialize final_list */
  PRISM_final_list_create( &final_list );

  /* Generate subproblems for final_list */
  if (me==ROOT) printf("\nstart PRISM_divide\n");
  t_start = MPI_Wtime();
  PRISM_divide(A, Z, D, cur_est, cut_off, final_list);
  t_syisda[t_test] = (MPI_Wtime()-t_start);

  /* Solve subproblems in final_list */
  if (me==ROOT) printf("start PRISM_end_game\n");
  t_start = MPI_Wtime();
  PRISM_end_game(final_list); 
  t_syisda[t_end] += (MPI_Wtime()-t_start);
  t_syisda[t_test] += t_syisda[t_end];

  /* Free objects */
  PLA_Obj_free(&cur_est);
  PRISM_final_list_delete(&final_list);
}





